import numpy as np
from numba import njit
from ._numba_settings import numba_default
from ._tools import atleast_1d
from .lineardispersion import intrinsic_dispersion_relation
from .constants import GRAV, WATER_DENSITY


@njit(**numba_default)
def horizontal_particle_velocity_amplitude(surface_amplitude, k, z, depth, surface_elevation=0, grav=GRAV):
    """
    :param surface_amplitude: Surface amplitude (m)
    :param k: Wavenumber (rad/m)
    :param z: Depth (m)
    :param depth: Depth (m)
    :param direction: Direction (rad)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    k = atleast_1d(k)
    z = atleast_1d(z)
    angular_frequency = intrinsic_dispersion_relation(k, depth, grav=grav)

    s = s_coordinate(z, depth, surface_elevation)
    kd = k * depth
    ch = np.cosh(kd * (1 + s)) / np.cosh(kd)

    return surface_amplitude * ch * grav * k / angular_frequency


@njit(**numba_default)
def vertical_particle_velocity_amplitude(surface_amplitude, k, z, depth, surface_elevation=0, grav=GRAV):
    """
    :param surface_amplitude: Surface amplitude (m)
    :param k: Wavenumber (rad/m)
    :param z: Depth (m)
    :param depth: Depth (m)
    :param direction: Direction (rad)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    k = atleast_1d(k)
    z = atleast_1d(z)
    angular_frequency = intrinsic_dispersion_relation(k, depth, grav=grav)

    s = s_coordinate(z, depth, surface_elevation)
    kd = k * depth
    sh = np.sinh(kd * (1 + s)) / np.cosh(kd)

    return surface_amplitude * sh * grav * k / angular_frequency


@njit(**numba_default)
def particle_velocity_amplitude_x(surface_amplitude, direction, k, z, depth, surface_elevation=0,
                                  grav=GRAV):
    """
    :param surface_amplitude: Surface amplitude (m)
    :param k: Wavenumber (rad/m)
    :param z: Depth (m)
    :param depth: Depth (m)
    :param direction: Direction (rad)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """

    return horizontal_particle_velocity_amplitude(surface_amplitude, k, z, depth, direction=direction,
                                                  surface_elevation=surface_elevation, grav=grav) * np.cos(direction)


@njit(**numba_default)
def particle_velocity_amplitude_y(surface_amplitude, direction, k, z, depth, surface_elevation=0,
                                  grav=GRAV):
    """
    :param surface_amplitude: Surface amplitude (m)
    :param k: Wavenumber (rad/m)
    :param z: Depth (m)
    :param depth: Depth (m)
    :param direction: Direction (rad)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """

    return horizontal_particle_velocity_amplitude(surface_amplitude, k, z, depth, direction=direction,
                                                  surface_elevation=surface_elevation, grav=grav) * np.sin(direction)


@njit(**numba_default)
def particle_velocity_amplitude_z(surface_amplitude, k, z, depth, surface_elevation=0, grav=GRAV):
    """
    :param surface_amplitude: Surface amplitude (m)
    :param k: Wavenumber (rad/m)
    :param z: Depth (m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """

    return -1j * vertical_particle_velocity_amplitude(surface_amplitude, k, z, depth,
                                                      surface_elevation=surface_elevation, grav=grav)


def pressure_amplitude(surface_amplitude, k, z, depth, surface_elevation=0, grav=GRAV, density=WATER_DENSITY):
    """
    :param surface_amplitude: Surface amplitude (m)
    :param k: Wavenumber (rad/m)
    :param z: Depth (m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    k = atleast_1d(k)
    z = atleast_1d(z)
    s = s_coordinate(z, depth, surface_elevation)
    ch = np.cosh(k * depth * (1 + s)) / np.cosh(k * depth)

    return density * grav * surface_amplitude * (ch - (depth + z) / (depth + surface_elevation))


@njit(**numba_default)
def s_coordinate(z, depth, surface_elevation):
    """
    :param k: Wavenumber (rad/m)
    :param z: Depth (m)
    :param depth: Depth (m)
    :param direction: Direction (rad)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    z = atleast_1d(z)
    return z - surface_elevation / (depth + surface_elevation)



import numpy as np
from numba import njit, types
from numba.extending import overload

# The following overloading trick is needed because "atleast_1d" is not supported for scalars by default in numba.
def atleast_1d(x) -> np.ndarray:
    if type(x) in types.number_domain:
        return np.array([x])
    return np.atleast_1d(x)

@overload(atleast_1d)
def overloaded_atleast_1d(x):
    if x in types.number_domain:
        return lambda x: np.array([x])
    return lambda x: np.atleast_1d(x)

def atleast_2d(x) -> np.ndarray:
    if x in types.number_domain:
        return np.array([x])
    return np.atleast_1d(x)

@overload(atleast_2d)
def overloaded_atleast_2d(x):
    if x in types.number_domain:
        return lambda x: np.array([[x]])
    return lambda x: np.atleast_2d(x)


from .spectrum import FrequencyDirectionSpectrum, create_2d_spectrum, FrequencySpectrum
from abc import ABC, abstractmethod
from scipy.special import gamma
from datetime import datetime
import numpy
from typing import Literal

PHILLIPS_CONSTANT = 0.0081
GRAVITATIONAL_CONSTANT = 9.81

FrequencyShapeOptions = Literal["pm", "jonswap", "phillips", "gaussian"]
DirectionalShapeOptions = Literal["raised_cosine"]


class FrequencyShape(ABC):
    @abstractmethod
    def values(self, frequency_hertz: numpy.ndarray) -> numpy.ndarray:
        pass


class DirectionalShape(ABC):
    @abstractmethod
    def values(self, direction_degrees: numpy.ndarray) -> numpy.ndarray:
        pass


class RaisedCosine(DirectionalShape):
    def __init__(self, mean_direction_degrees: float = 0, width_degrees: float = 28.64):
        self._power = self.power(width_degrees)
        self._mean_direction_degrees = mean_direction_degrees
        self._normalization = (
            numpy.pi
            / 180
            * gamma(self._power / 2 + 1)
            / (gamma(self._power / 2 + 1 / 2) * numpy.sqrt(numpy.pi))
        )

    @staticmethod
    def power(width_degrees):
        return 4 / ((numpy.pi * width_degrees / 90) ** 2) - 2

    def values(self, direction_degrees: numpy.ndarray) -> numpy.ndarray:
        angle = (direction_degrees - self._mean_direction_degrees + 180) % 360 - 180
        with numpy.errstate(invalid="ignore", divide="ignore"):
            return numpy.where(
                numpy.abs(angle) <= 90,
                self._normalization * numpy.cos(angle * numpy.pi / 180) ** self._power,
                0,
            )


class GaussianSpectrum(FrequencyShape):
    def __init__(self, peak_frequency_hertz, m0: float = 1, **kwargs):
        self.m0 = m0
        self._peak_frequency_hertz = peak_frequency_hertz
        self.standard_deviation_hertz = kwargs.get(
            "standard_deviation_hertz", peak_frequency_hertz / 10
        )

    def values(self, frequency_hertz: numpy.ndarray) -> numpy.ndarray:
        return (
            self.m0
            / self.standard_deviation_hertz
            / numpy.sqrt(2 * numpy.pi)
            * numpy.exp(
                -0.5
                * (frequency_hertz - self._peak_frequency_hertz) ** 2
                / self.standard_deviation_hertz**2
            )
        )


class PhillipsSpectrum(FrequencyShape):
    def __init__(self, peak_frequency_hertz, m0: float = 1, **kwargs):

        self._peak_frequency_hertz = peak_frequency_hertz
        self._g = kwargs.get("g", GRAVITATIONAL_CONSTANT)
        self._alpha = self.alpha(m0)

    def alpha(self, m0):
        return m0 * 8 * (numpy.pi) ** 4 * self._peak_frequency_hertz**4 / self._g**2

    def values(self, frequency_hertz: numpy.ndarray) -> numpy.ndarray:
        """
        Phillips variance-density spectrum with frequency in Hz as
        dependent variable.

        :return:
        """
        values = numpy.zeros(len(frequency_hertz))
        msk = frequency_hertz > 0
        values[msk] = (
            self._alpha
            * self._g**2
            * (2 * numpy.pi) ** -4
            * frequency_hertz[msk] ** -5
        )
        return values


class PiersonMoskowitzSpectrum(FrequencyShape):
    def __init__(self, peak_frequency_hertz, m0: float = 1, **kwargs):
        self._peak_frequency_hertz = peak_frequency_hertz
        self._g = kwargs.get("g", GRAVITATIONAL_CONSTANT)
        self._alpha = self.alpha(m0)

    def alpha(self, m0):
        return m0 * 5 * (2 * numpy.pi * self._peak_frequency_hertz) ** 4 / self._g**2

    def values(self, frequency_hertz: numpy.ndarray) -> numpy.ndarray:
        """
        Pierson Moskowitz variance-density spectrum with frequency in Hz as
        dependant variable. See e.g. Holthuijsen "Waves in Oceanic Water."

        :param frequency: frequency in Hz (scalar or array)
        :param peak_frequency: peak frequency in Hz
        :param alpha: Phillips constant (default 0.0081)
        :param g: gravitational acceleration (default 9.81)
        :return:
        """
        values = numpy.zeros(len(frequency_hertz))
        msk = frequency_hertz > 0
        values[msk] = (
            self._alpha
            * self._g**2
            * (2 * numpy.pi) ** -4
            * frequency_hertz[msk] ** -5
            * numpy.exp(
                -5 / 4 * (self._peak_frequency_hertz / frequency_hertz[msk]) ** 4
            )
        )
        return values


class JonswapSpectrum(FrequencyShape):
    def __init__(self, peak_frequency_hertz, m0: float = 1, **kwargs):
        self._peak_frequency_hertz = peak_frequency_hertz
        self._g = kwargs.get("g", GRAVITATIONAL_CONSTANT)
        self._sigma_a = kwargs.get("sigma_a", 0.07)
        self._sigma_b = kwargs.get("sigma_b", 0.09)
        self.gamma = kwargs.get("gamma", 3.3)
        self._alpha = self.alpha(m0)

    def alpha(self, m0):
        # Approximation by Yamaguchi (1984), "Approximate expressions for integral properties of the JONSWAP
        # spectrum" Proc. Japanese Society of Civil Engineers, 345/II-1, 149–152 [in Japanese]. Taken from Holthuijsen
        # "waves in oceanic and coastal waters". Not valid if sigma_a or sigma_b are chanegd from defaults. Otherwise
        # accurate to within 0.25%
        #
        return (
            m0
            * (2 * numpy.pi * self._peak_frequency_hertz) ** 4
            / self._g**2
            / (0.06533 * self.gamma**0.8015 + 0.13467)
        )

    def values(self, frequency_hertz: numpy.ndarray) -> numpy.ndarray:
        """
        Jonswap variance-density spectrum with frequency in Hz as
        dependant variable. See e.g. Holthuijsen "Waves in Oceanic Water."

        :param frequency: frequency in Hz (scalar or array)
        :param peak_frequency: peak frequency in Hz
        :param alpha: Phillips constant (default 0.0081)
        :param g: gravitational acceleration (default 9.81)
        :return:
        """
        values = numpy.zeros(len(frequency_hertz))
        msk = frequency_hertz > 0

        sigma = numpy.where(
            frequency_hertz <= self._peak_frequency_hertz, self._sigma_a, self._sigma_b
        )
        peak_enhancement = self.gamma ** numpy.exp(
            -1 / 2 * ((frequency_hertz / self._peak_frequency_hertz - 1) / sigma) ** 2
        )

        values[msk] = (
            self._alpha
            * self._g**2
            * (2 * numpy.pi) ** -4
            * frequency_hertz[msk] ** -5
            * numpy.exp(
                -5 / 4 * (self._peak_frequency_hertz / frequency_hertz[msk]) ** 4
            )
            * peak_enhancement[msk]
        )
        return values


def create_frequency_shape(
    shape: FrequencyShapeOptions, peak_frequency_hertz: float, m0: float = 1, **kwargs
) -> FrequencyShape:
    if shape == "pm":
        return PiersonMoskowitzSpectrum(
            peak_frequency_hertz=peak_frequency_hertz, m0=m0, **kwargs
        )
    elif shape == "jonswap":
        return JonswapSpectrum(
            peak_frequency_hertz=peak_frequency_hertz, m0=m0, **kwargs
        )
    elif shape == "phillips":
        return PhillipsSpectrum(
            peak_frequency_hertz=peak_frequency_hertz, m0=m0, **kwargs
        )
    elif shape == "gaussian":
        return GaussianSpectrum(
            peak_frequency_hertz=peak_frequency_hertz, m0=m0, **kwargs
        )

    else:
        raise ValueError(f"Unknown frequency shape: {shape}")


def create_directional_shape(
    shape: DirectionalShapeOptions,
    mean_direction_degrees: float = 0,
    width_degrees: float = 30,
) -> DirectionalShape:
    if shape == "raised_cosine":
        return RaisedCosine(
            mean_direction_degrees=mean_direction_degrees, width_degrees=width_degrees
        )
    else:
        raise ValueError(f"Unknown frequency shape: {shape}")


def create_parametric_frequency_direction_spectrum(
    frequency_hertz: numpy.ndarray,
    peak_frequency_hertz: float,
    significant_wave_height: float,
    frequency_shape: FrequencyShapeOptions = "jonswap",
    direction_degrees: numpy.ndarray = None,
    direction_shape: DirectionalShapeOptions = "raised_cosine",
    mean_direction_degrees: float = 0.0,
    width_degrees: float = 30,
    depth=numpy.inf,
    time: datetime = None,
    latitude: float = None,
    longitude: float = None,
    **kwargs,
) -> FrequencyDirectionSpectrum:
    """
    Create a parametrized directional frequency spectrum according to a given frequency (Jonswap, PM) or directional
    (raised_cosine) distribution.

    :param frequency_hertz: Frequencies to resolve
    :param peak_frequency_hertz:  Desired peak frequency of the spectrum
    :param significant_wave_height: Significant wave height of the spectrum
    :param frequency_shape: The frequency shape, currently supported are:
        frequency_shape="pm": for pierson_moskowitz
        frequency_shape="jonswap" [default]: for Jonswap
    :param direction_degrees: Directions to resolve the spectrum. If None [default] 36 directions spanning the circle
        are used [ 0 , 360 )
    :param direction_shape: shape of the directional distribution. Currently only a raised cosine distribution is
        supported.
    :param mean_direction_degrees: mean direction of the waves. 0 degrees (due east) is the default.
    :param width_degrees: width of the spectrum (according to Kuik). 30 degrees is the default.
    :param depth: mean depth at the location of the spectrum (optional). Does not affect returned spectral values in any
        way, but is used as the depth in the returned spectral object (and may affect e.g. wavenumber calculations.)
    :param time: timestamp of the spectrum. Optional. Merely an annotation on the returned object.
    :param latitude: latitude of the spectrum. Optional. Merely an annotation on the returned object.
    :param longitude: latitude of the spectrum. Optional. Merely an annotation on the returned object.

    :return: FrequencyDirectionSpectrum object.
    """

    if direction_degrees is None:
        direction_degrees = numpy.linspace(0, 360, 36, endpoint=False)

    D = create_directional_shape(
        shape=direction_shape,
        mean_direction_degrees=mean_direction_degrees,
        width_degrees=width_degrees,
    ).values(direction_degrees)

    m0 = (significant_wave_height / 4) ** 2
    E = create_frequency_shape(
        shape=frequency_shape,
        peak_frequency_hertz=peak_frequency_hertz,
        m0=m0,
        **kwargs,
    ).values(frequency_hertz)

    return create_2d_spectrum(
        frequency=frequency_hertz,
        direction=direction_degrees,
        variance_density=E[:, None] * D[None, :],
        time=time,
        latitude=latitude,
        longitude=longitude,
        depth=depth,
        dims=("frequency", "direction"),
    )


def create_parametric_frequency_spectrum(
    frequency_hertz: numpy.ndarray,
    peak_frequency_hertz: float,
    significant_wave_height: float,
    frequency_shape: FrequencyShapeOptions = "jonswap",
    depth=numpy.inf,
    time: datetime = None,
    latitude: float = None,
    longitude: float = None,
    **kwargs,
) -> FrequencySpectrum:

    # We create a 1d spectrum from an integrated 2d spectrum with assumed raised cosine shape. This allows us to
    # add the a1/b1 parameters easily.
    spec2d = create_parametric_frequency_direction_spectrum(
        frequency_hertz,
        peak_frequency_hertz,
        significant_wave_height,
        frequency_shape,
        depth=depth,
        time=time,
        latitude=latitude,
        longitude=longitude,
        **kwargs,
    )
    return spec2d.as_frequency_spectrum()


def create_parametric_spectrum(
    frequency_hertz: numpy.ndarray,
    frequency_shape: FrequencyShapeOptions,
    peak_frequency_hertz: float,
    significant_wave_height: float,
    direction_degrees: numpy.ndarray = None,
    direction_shape: DirectionalShapeOptions = "raised_cosine",
    mean_direction_degrees: float = 0.0,
    width_degrees: float = 30.0,
    depth=numpy.inf,
    time: datetime = None,
    latitude: float = None,
    longitude: float = None,
) -> FrequencyDirectionSpectrum:
    """
    Deprecated - use create_parametric_frequency_direction_spectrum instead
    """
    return create_parametric_frequency_direction_spectrum(
        frequency_hertz,
        peak_frequency_hertz,
        significant_wave_height,
        frequency_shape,
        direction_degrees,
        direction_shape,
        mean_direction_degrees,
        width_degrees,
        depth,
        time,
        latitude,
        longitude,
    )



"""
Contents: Routines to generate a (monotone) cubic spline interpolation for 1D arrays.

Copyright (C) 2023
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Functions:

- `cubic_spline`, method to create a (monotone) cubic spline

"""
import numpy as np
from scipy.interpolate import CubicSpline

try:
    from qpsolvers import solve_ls
except ImportError:
    pass

def cubic_spline(
    x: np.ndarray,
    y: np.ndarray,
    monotone_interpolation: bool = False,
    frequency_axis=-1,
) -> CubicSpline:
    """
    Construct a cubic spline, optionally monotone.

    :param x: array_like, shape (n,)
              1-D array containing values of the independent variable.
              Values must be real, finite and in strictly increasing order.
    :param y: array_like, shape (...,n)
              set of m 1-D arrays containing values of the dependent variable. Y can have an arbitrary set of leading
              dimensions, but the last dimension has the be equal in size to X. Values must be real, finite and in
              strictly increasing order along the last dimension. (Y is assumed monotone).

    :param monotone_interpolation:
    :return:
    """

    if not monotone_interpolation:
        spline = CubicSpline(x, y, axis=frequency_axis)

    else:
        # Reshape Y, which can have arbitrary leading dimensions (including none) - into a (m,n) shape.
        input_shape = y.shape
        input_axis = list(range(len(input_shape)))
        frequency_axis = input_axis[frequency_axis]

        if len(input_shape) == 1:
            shape = (1, len(x))
            Y = y
        else:
            shape = (
                np.prod(input_shape) // input_shape[frequency_axis],
                input_shape[frequency_axis],
            )
            axis = [axis for axis in input_axis if axis != frequency_axis]
            axis.append(frequency_axis)
            Y = np.transpose(y, axis)
            permuted_input_shape = Y.shape
        Y = np.reshape(Y, shape)

        # Create the spline coeficients
        output = monotone_cubic_spline_coeficients(x, Y)

        # Reshape output so that the aribrary leading dimensions of y become _trailing_ dimensions of the output
        # (this is how spline coeficients are stored in the scipy CubicSpline object)
        if len(input_shape) == 1:
            output_shape = (4, input_shape[-1] - 1)
            output = output.reshape(output_shape)
        else:
            output_shape = (
                *permuted_input_shape[0:-1],
                4,
                permuted_input_shape[-1] - 1,
            )
            output = output.reshape(output_shape)
            output_axis = list(range(len(output_shape)))
            output_axis = output_axis[-2:] + output_axis[:-2]
            output = np.transpose(output, output_axis)

        # Return a CubicSpline object.
        spline = CubicSpline.construct_fast(output, x, extrapolate=False, axis=1)

    return spline


def monotone_cubic_spline_coeficients(x: np.ndarray, Y: np.ndarray) -> np.ndarray:
    """
    Construct the spline coeficients.
    :param x: array_like, shape (n,)
              1-D array containing values of the independent variable.
              Values must be real, finite and in strictly increasing order.
    :param Y: array_like, shape (m,n)
              set of m 1-D arrays containing values of the dependent variable. For each of the m rows an independent
              spline will be constructed. Values must be real, finite and in strictly increasing order. (Y is assumed
              monotone).
    :param monotone:
    :return:
    """

    output = np.zeros((Y.shape[0], 4, Y.shape[1] - 1))

    # loop over the dimensions
    for jj in range(0, Y.shape[0]):
        _ = _monotone_cubic_spline(x, Y[jj, :], output[jj, :, :])

    return output


def _monotone_cubic_spline(
    x: np.ndarray, y: np.ndarray, spline_coeficients=None
) -> np.ndarray:
    """
    Find a monotone cubic spline function that is maximally smooth. The basic idea is that instead of the traditional
    natural spline - which has C2 continuity but is not guaranteed monotone we reformulate the spline problem as a
    constrained minimization, where the requirements of C0 and C1 continuity and the monoticity conditions form the
    (in)equality constraints, and we try to find a curve that meets these constraints, but is otherwise as close to
    C2 continuity as possible. The spline boundary conditions are not-a-knot boundary conditions, which are added to
    the minimization problem (not as a constraint).

        Wolberg, G., & Alfy, I. (1999, June). Monotonic cubic spline interpolation.
        In Computer Graphics International (pp. 188-195).

        minimize ::

                (matrix @ X - rhs) @ (matrix @ X - rhs)^T

        such that

                equality_constraint_matrix @ X = equality_constraint_rhs
                inequality_constraint_matrix @ X <= inequality_constraint_rhs

    The matrix in the least-squares minimization is formed by minimizing the discontinuity in the second-derivative at
    the spline edges. Note: that because our solver (quadprog) cannot handle matrices that are rank-deficient, we add
    the equality constraints to the minimization problem. This does not alter the result- but allows the solver to
    proceed. (changing solver in future is desirable)

    The equality constraints are formed from the spline conditions that the curve and its first derivative are
    continuous, while we try to minimize the discontinuity in the second derivative at spline edges.

    The inequality constraints are formed from the monoticity conditions (see Wolfberg, 1999).

    :param x: array_like, shape (n,)
              1-D array containing values of the independent variable.
              Values must be real, finite and in strictly increasing order.
    :param y: array_like, shape (n,)
              1-D array containing values of the dependent variable. For each of the m rows an independent
              spline will be constructed. Values must be real, finite and in strictly increasing order. (y is assumed
              monotone).
    :param spline_coeficients: (optional) array_like, shape (4,n)
              Output array to change values in place.
    :return: array_like, shape (4, (n-1))
    """

    # Initialize arrays
    # -----------------
    if spline_coeficients is None:
        spline_coeficients = np.zeros((4, len(x) - 1))

    number_of_splines = x.shape[-1] - 1
    least_squares_matrix = np.zeros((3 * (number_of_splines), 3 * number_of_splines))
    least_squares_rhs = np.zeros((3 * (number_of_splines),))

    equality_constraint_matrix = np.zeros(
        (2 * (number_of_splines - 1) + 1, 3 * number_of_splines)
    )
    equality_constraint_rhs = np.zeros(2 * (number_of_splines - 1) + 1)

    inequality_constraint_matrix = np.zeros(
        (4 * (number_of_splines), 3 * number_of_splines)
    )
    inequality_constraint_rhs = np.zeros((4 * (number_of_splines)))

    delta_x = np.diff(x)
    delta_x = np.reshape(delta_x, (len(delta_x), 1))

    delta_y = np.diff(y)
    secant = np.diff(y) / np.diff(x)
    max_curvature = np.max(
        np.abs(2 * np.diff(y, n=2) / (np.diff(x)[0:-1] + np.diff(x)[1:]))
    )  # 2 * np.diff( y,n=2 ) / (np.diff(x)[0:-1] + np.diff(x)[1:])
    max_secant = np.max(
        np.abs(secant)
    )  # [ np.max( (secant[ii],secant[ii+1]) ) for ii in range(0,len(secant)-1) ]
    max_delta = np.max(np.abs(delta_y))

    if max_delta == 0:
        # Zero solution
        return spline_coeficients

    ones = np.ones_like(delta_x)
    zeros = np.zeros_like(delta_x)
    twos = np.full_like(delta_x, 2)

    # these are the coeficients that represent the linear equations for C0, C1 and C2 continuity at spline edges.
    d2y_dx2_coef = np.concatenate((6 * delta_x, twos, zeros, zeros, -twos), axis=1)
    dy_dx_coef = np.concatenate(
        (3 * delta_x**2, 2 * delta_x, ones, zeros, zeros, -ones), axis=1
    )
    y_coef = np.concatenate((delta_x**3, delta_x**2, delta_x), axis=1)

    # Create Matrices
    # -----------------

    # Not-a-knot boundary condition.
    least_squares_matrix[0, 0] = 1
    least_squares_matrix[0, 3] = -1
    least_squares_matrix[-1, number_of_splines * 3 - 6] = 1
    least_squares_matrix[-1, number_of_splines * 3 - 3] = -1

    # loop over each of the splines, and add its equations
    weights = [1, 10000, 10000]
    for ii in range(0, number_of_splines - 1):
        # row/column indices in the matrix. The unknown spline coeficients are stored as a linear vector of the form
        # solution = [ a0,b0,c0; a1,b1,c1; .... ]. So the spline coeficients of the ii'th spline start in the
        # ii*3'th column
        jcol = ii * 3
        jrow = ii * 3 + 1

        # d2y/dx2 continuity at spline edges. Note technically this is the only equation we minimize
        least_squares_matrix[jrow, jcol : jcol + 5] = (
            d2y_dx2_coef[ii, :] / max_curvature * weights[0]
        )

        # dy/dx continuity at spline edges
        least_squares_matrix[jrow + 1, jcol : jcol + 6] = (
            dy_dx_coef[ii, :] / max_secant * weights[1]
        )

        # y continuity at spline edges
        least_squares_matrix[jrow + 2, jcol : jcol + 3] = (
            y_coef[ii, :] / max_delta * weights[2]
        )
        least_squares_rhs[jrow + 2] = delta_y[ii] / max_delta * weights[2]

        # build the matrix with the equality constrains
        jrow = ii * 2

        # # dy/dx continuity at spline edges.
        if secant[ii] * secant[ii + 1] > 0.0:
            # Note we skip this constraint if we are at a peak in the data- or if one of the surrounding secants is 0.
            # Enforcing 0.0 slope (the other posibility in this case) can lead to a system of equations that is not
            # solvable.
            equality_constraint_matrix[jrow, jcol : jcol + 6] = dy_dx_coef[ii, :]
        else:
            equality_constraint_matrix[jrow, jcol : jcol + 3] = dy_dx_coef[ii, :3]

        equality_constraint_matrix[jrow + 1, jcol : jcol + 3] = y_coef[ii, :]
        equality_constraint_rhs[jrow + 1] = delta_y[ii]

    # Add the end node values for the final spline
    equality_constraint_matrix[-1, -3:] = y_coef[-1, :]
    equality_constraint_rhs[-1] = delta_y[-1]

    least_squares_matrix[-2, number_of_splines * 3 - 3 :] = y_coef[-1, :] / max_delta
    least_squares_rhs[-2] = delta_y[-1] / max_delta

    # Build the matrix for the inequality constraints. These constraints enfore monotone behaviour of each spline.
    eps = 0.0  # 1e-10
    for ii in range(0, number_of_splines):
        jcol = ii * 3
        jrow = ii * 4

        if ii > 0:
            check = secant[ii] * secant[ii - 1] > 0
        else:
            check = np.abs(secant[ii]) > 0

        sign = 1.0 if secant[ii] >= 0 else -1.0

        if check:
            # dy/dx start of spline must be smaller than 3 times the secant
            inequality_constraint_matrix[jrow, jcol + 2] = sign * 1.0
            inequality_constraint_rhs[jrow] = 3 * secant[ii]

            # dy/dx start of spline must be of the same sign as the secant
            inequality_constraint_matrix[jrow + 2, jcol + 2] = -sign
            inequality_constraint_rhs[jrow + 2] = eps
        else:
            # secants change sign across node - slope must be zero-> slope <= 0  & -slope <=0
            inequality_constraint_matrix[jrow, jcol + 2] = 1.0
            inequality_constraint_rhs[jrow] = eps

            inequality_constraint_matrix[jrow + 2, jcol + 2] = -1
            inequality_constraint_rhs[jrow + 2] = eps

        if ii < number_of_splines - 1:
            check = secant[ii + 1] * secant[ii] > 0
        else:
            check = np.abs(secant[ii]) > 0

        if check:
            # dy/dx at the end of the spline must be smaller than 3 times the secant
            inequality_constraint_matrix[jrow + 1, jcol : jcol + 3] = (
                sign * dy_dx_coef[ii, :3]
            )
            inequality_constraint_rhs[jrow + 1] = 3 * secant[ii]

            # dy/dx at the end of the spline must be larger than 0
            inequality_constraint_matrix[jrow + 3, jcol : jcol + 3] = (
                -sign * dy_dx_coef[ii, :3]
            )
            inequality_constraint_rhs[jrow + 3] = eps
        else:
            # secants change sign across node - slope must be zero-> slope <= 0  & -slope <=0
            inequality_constraint_matrix[jrow + 1, jcol : jcol + 3] = dy_dx_coef[ii, :3]
            inequality_constraint_rhs[jrow + 1] = eps

            inequality_constraint_matrix[jrow + 3, jcol : jcol + 3] = -dy_dx_coef[
                ii, :3
            ]
            inequality_constraint_rhs[jrow + 3] = eps

    # Solve system and return results
    # -----------------

    # Solve the solution as a constrained least squares minimization
    solution = solve_ls(
        R=least_squares_matrix,
        s=least_squares_rhs,
        G=inequality_constraint_matrix,
        h=inequality_constraint_rhs,
        A=equality_constraint_matrix,
        b=equality_constraint_rhs,
        verbose=False,
        solver="cvxopt",
    )

    # The unknown spline coeficients are stored as a linear vector of the form:
    #
    #    solution = [ a0,b0,c0; a1,b1,c1; .... ].
    #
    # here we unpack that.
    if solution is None:
        # Remove the dydx continuity requirement.. as a last resort
        equality_constraint_rhs = equality_constraint_rhs[1::2]
        equality_constraint_matrix = equality_constraint_matrix[1::2, :]
        solution = solve_ls(
            R=least_squares_matrix,
            s=least_squares_rhs,
            G=inequality_constraint_matrix,
            h=inequality_constraint_rhs,
            A=equality_constraint_matrix,
            b=equality_constraint_rhs,
            verbose=False,
            solver="cvxopt",
        )
        print(
            "warning - no monotone solution found attempting to solve without C1 constraint"
        )
        if solution is None:
            print("warning - no monotone solution found")
            raise Exception("No solution found.")

    solution = np.reshape(solution, (number_of_splines, 3))

    spline_coeficients[0, :] = solution[:, 0]  # "a" coef
    spline_coeficients[1, :] = solution[:, 1]  # "b" coef
    spline_coeficients[2, :] = solution[:, 2]  # "c" coef
    spline_coeficients[3, :] = y[:-1]  # "d

    #  return the spline coeficients
    return spline_coeficients



from typing import Tuple

from numba import jit
from numpy import inf, cos, sin, sqrt, arctan2, pi, exp, isnan, nan, empty, log, any
from numpy.typing import NDArray

from roguewave.wavephysics.balance.wam_tail_stress import _charnock_relation_point
from roguewave.wavephysics.balance.solvers import numba_newton_raphson
from roguewave.wavetheory import inverse_intrinsic_dispersion_relation
from roguewave.wavephysics.balance._numba_settings import numba_nocache


@jit(**numba_nocache)
def _wave_supported_stress_point(
    wind_input: NDArray,
    depth,
    spectral_grid,
    variance_density,
    wind,
    roughness_length,
    tail_stress_parametrization_function,
    parameters,
) -> Tuple[float, float]:
    """
    :param wind_input:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """

    number_of_frequencies, number_of_directions = wind_input.shape

    radian_frequency = spectral_grid["radian_frequency"]
    radian_direction = spectral_grid["radian_direction"]
    frequency_step = spectral_grid["frequency_step"]
    direction_step = spectral_grid["direction_step"]
    gravitational_acceleration = parameters["gravitational_acceleration"]

    if depth == inf:
        wavenumber = radian_frequency**2 / gravitational_acceleration
    else:
        wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)

    cosine_step = cos(radian_direction) * direction_step
    sine_step = sin(radian_direction) * direction_step

    # Calculate the resolved part of the wavestress over the given input spectrum
    stress_east = 0.0
    stress_north = 0.0
    for frequency_index in range(number_of_frequencies):
        inverse_wave_speed = (
            wavenumber[frequency_index] / radian_frequency[frequency_index]
        ) * frequency_step[frequency_index]

        for direction_index in range(number_of_directions):
            stress_east += (
                cosine_step[direction_index]
                * wind_input[frequency_index, direction_index]
                * inverse_wave_speed
            )

            stress_north += (
                sine_step[direction_index]
                * wind_input[frequency_index, direction_index]
                * inverse_wave_speed
            )

    # calculate the magnitude
    stress_north *= (
        parameters["gravitational_acceleration"] * parameters["water_density"]
    )
    stress_east *= (
        parameters["gravitational_acceleration"] * parameters["water_density"]
    )

    # Add stress contribution due to unresolved waves in the spectral tail (above the last resolved frequency)
    (
        eastward_tail_wave_stress,
        northward_tail_wave_stress,
    ) = tail_stress_parametrization_function(
        variance_density, wind, depth, roughness_length, spectral_grid, parameters
    )
    stress_east += eastward_tail_wave_stress
    stress_north += northward_tail_wave_stress

    # Total stress is the sum or resolved and unresolved part.
    return stress_east, stress_north  # wave_stress_magnitude, wave_stress_direction


@jit(**numba_nocache)
def _roughness_estimate_point(
    guess,
    variance_density,
    wind,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    spectral_grid,
    parameters,
):
    """

    :param guess:
    :param variance_density:
    :param wind:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """

    vonkarman_constant = parameters["vonkarman_constant"]
    if guess < 0:
        if wind[2] == "u10":
            drag_coeficient_wu = (0.8 + 0.065 * wind[0]) / 1000
            guess = 10 / exp(vonkarman_constant / sqrt(drag_coeficient_wu))

        elif wind[2] in ["ustar", "friction_velocity"]:
            guess = _charnock_relation_point(wind[0], parameters)
        else:
            raise ValueError("unknown wind forcing")

    if any(isnan(variance_density)):
        return nan

    if wind[0] == 0.0:
        return nan

    generation = empty((variance_density.shape))

    function_arguments = (
        variance_density,
        wind,
        depth,
        wind_source_term_function,
        tail_stress_parametrization_function,
        spectral_grid,
        parameters,
        generation,
    )

    log_root = numba_newton_raphson(
        _stress_iteration_function,
        log(guess),
        function_arguments,
        hard_bounds=(-20, 0),
        relative_stepsize=False,
        atol=1e-6,
        rtol=1e-6,
        error_on_max_iter=True,
        max_iterations=100,
        verbose=False,
        name="stress iteration",
        aitken_acceleration=False,
    )
    return exp(log_root)


@jit(**numba_nocache)
def _roughness_estimate(
    guess,
    variance_density,
    wind,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    spectral_grid,
    parameters,
) -> NDArray:
    """
    :param guess:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """
    number_of_points = variance_density.shape[0]
    roughness = empty((number_of_points))

    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])

        if isnan(wind[0][point_index]):
            roughness[point_index] = nan
        else:
            try:
                roughness[point_index] = _roughness_estimate_point(
                    guess=guess[point_index],
                    variance_density=variance_density[point_index, :, :],
                    wind=wind_at_point,
                    depth=depth[point_index],
                    wind_source_term_function=wind_source_term_function,
                    tail_stress_parametrization_function=tail_stress_parametrization_function,
                    spectral_grid=spectral_grid,
                    parameters=parameters,
                )
            except:
                roughness[point_index] = nan
    return roughness


@jit(**numba_nocache)
def _wave_supported_stress(
    variance_density,
    wind,
    depth,
    roughness_length,
    wind_source_term_function,
    tail_stress_parametrization_function,
    spectral_grid,
    parameters,
) -> Tuple[NDArray, NDArray]:
    """
    Calculate the wave supported wind stress.

    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param roughness_length: Surface roughness length in meter.
    :param spectral_grid:
    :param parameters:
    :return:
    """
    number_of_points = variance_density.shape[0]
    magnitude = empty((number_of_points))
    direction = empty((number_of_points))
    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])
        (magnitude[point_index], direction[point_index],) = _total_stress_point(
            roughness_length[point_index],
            variance_density[point_index, :, :],
            wind=wind_at_point,
            depth=depth[point_index],
            wind_source_term_function=wind_source_term_function,
            tail_stress_parametrization_function=tail_stress_parametrization_function,
            spectral_grid=spectral_grid,
            parameters=parameters,
        )
    return magnitude, direction


@jit(**numba_nocache)
def _tail_supported_stress(
    variance_density,
    wind,
    depth,
    roughness_length,
    tail_stress_parametrization_function,
    spectral_grid,
    parameters,
) -> Tuple[NDArray, NDArray]:
    """
    Calculate the wave supported wind stress.

    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param roughness_length: Surface roughness length in meter.
    :param spectral_grid:
    :param parameters:
    :return:
    """
    number_of_points = variance_density.shape[0]
    stress_east = empty((number_of_points))
    stress_north = empty((number_of_points))
    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])
        (
            stress_east[point_index],
            stress_north[point_index],
        ) = tail_stress_parametrization_function(
            variance_density[point_index, :, :],
            wind_at_point,
            depth[point_index],
            roughness_length[point_index],
            spectral_grid,
            parameters,
        )

    magnitude = sqrt(stress_north**2 + stress_east**2)
    direction = arctan2(stress_north, stress_east) % 360
    return magnitude, direction


@jit(**numba_nocache)
def _stress_iteration_function(
    log_roughness_length,
    variance_density,
    wind,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    spectral_grid,
    parameters,
    work_array,
):
    """
    The surface roughness is defined implicitly. We use a fixed-point iteration step to solve for the roughness. This
    is the iteration function used in the fixed point iteration.

    :param roughness_length: Surface roughness length in meter.
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """
    roughness_length = exp(log_roughness_length)

    vonkarman_constant = parameters["vonkarman_constant"]
    elevation = parameters["elevation"]
    if wind[2] == "u10":
        friction_velocity = (
            wind[0] * vonkarman_constant / log(elevation / roughness_length)
        )
    else:
        friction_velocity = wind[0]

    # Get the wind input source term values
    work_array = wind_source_term_function(
        variance_density,
        wind,
        depth,
        roughness_length,
        spectral_grid,
        parameters,
        work_array,
    )

    total_stress_estimate, total_stress_direction_estimate = _total_stress_point(
        roughness_length,
        variance_density,
        wind,
        depth,
        wind_source_term_function,
        tail_stress_parametrization_function,
        spectral_grid,
        parameters,
        work_array,
    )

    return parameters["air_density"] * friction_velocity**2 - total_stress_estimate


@jit(**numba_nocache)
def _total_stress_point(
    roughness_length,
    variance_density,
    wind,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    spectral_grid,
    parameters,
    work_array=None,
):
    """

    :param roughness_length: Surface roughness length in meter.
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind: wind input tuple (magnitude, direction_degrees, type_string)
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """

    vonkarman_constant = parameters["vonkarman_constant"]
    elevation = parameters["elevation"]
    if wind[2] == "u10":
        friction_velocity = (
            wind[0] * vonkarman_constant / log(elevation / roughness_length)
        )
    else:
        friction_velocity = wind[0]

    if friction_velocity == 0.0:
        return 0.0, nan

    if isnan(friction_velocity):
        return nan, nan

    # Get the wind input source term values
    work_array = wind_source_term_function(
        variance_density,
        wind,
        depth,
        roughness_length,
        spectral_grid,
        parameters,
        work_array,
    )

    # Calculate the stress
    stress_east, stress_north = _wave_supported_stress_point(
        work_array,
        depth,
        spectral_grid,
        variance_density,
        wind,
        roughness_length,
        tail_stress_parametrization_function,
        parameters,
    )

    viscous_stress = (
        parameters["viscous_stress_parameter"]
        * parameters["air_density"]
        * friction_velocity
        * parameters["air_viscosity"]
        / vonkarman_constant
        / roughness_length
    )

    wind_direction_radian = wind[1] * pi / 180

    stress_north += viscous_stress * sin(wind_direction_radian)
    stress_east += viscous_stress * cos(wind_direction_radian)

    stress_direction = (arctan2(stress_north, stress_east) * 180 / pi) % 360
    stress_magnitude = sqrt(stress_north**2 + stress_east**2)

    return stress_magnitude, stress_direction



from roguewave.wavephysics.fluidproperties import (
    AIR,
    WATER,
    GRAVITATIONAL_ACCELERATION,
)

from roguewave.wavephysics.balance import WindGeneration
from roguewave.wavephysics.balance._numba_settings import numba_default
from numpy import cos, pi, log, exp, empty, inf
from numba import jit
from roguewave.wavetheory import inverse_intrinsic_dispersion_relation
from typing import TypedDict
from roguewave.wavephysics.balance.wam_tail_stress import (
    tail_stress_parametrization_wam,
)


class ST4WaveGenerationParameters(TypedDict):
    gravitational_acceleration: float
    charnock_maximum_roughness: float
    charnock_constant: float
    air_density: float
    water_density: float
    vonkarman_constant: float
    wave_age_tuning_parameter: float
    growth_parameter_betamax: float
    elevation: float
    air_viscosity: float
    viscous_stress_parameter: float


class ST4WindInput(WindGeneration):
    name = "st4 generation"

    def __init__(self, parameters: ST4WaveGenerationParameters = None):
        super(ST4WindInput, self).__init__(parameters)
        self._wind_source_term_function = _st4_wind_generation_point
        self._tail_stress_parametrization_function = tail_stress_parametrization_wam

    @staticmethod
    def default_parameters() -> ST4WaveGenerationParameters:
        return ST4WaveGenerationParameters(
            wave_age_tuning_parameter=0.006,
            growth_parameter_betamax=1.52,
            gravitational_acceleration=GRAVITATIONAL_ACCELERATION,
            charnock_maximum_roughness=inf,
            charnock_constant=0.01,
            air_density=AIR.density,
            water_density=WATER.density,
            vonkarman_constant=AIR.vonkarman_constant,
            elevation=10,
            air_viscosity=AIR.kinematic_viscosity,
            viscous_stress_parameter=0.0,
        )


# ----------------------------------------------------------------------------------------------------------------------
# ST4 Point wise implementation functions (apply to a single spatial point)
# ----------------------------------------------------------------------------------------------------------------------


@jit(**numba_default)
def _st4_wind_generation_point(
    variance_density,
    wind,
    depth: float,
    roughness_length: float,
    spectral_grid,
    parameters,
    wind_source=None,
):
    """
    Implementation of the st4 wind growth rate term as proposed by Janssen 1991- and as described in the ST4 paper.

    :param variance_density: input numpy array of shape (nf,nd), with nf number of frequencies, nd number of directions,
        that contains variance densities (unit m**2/Hz/Degree).
    :param wind: Wind input tuple (speed, direction, type) that descripes the wind speed, direction and kind of wind
        input. The latter can be: "u10" or "friction_velocity".
    :param depth: Local water depth in meter.
    :param roughness_length: Roughness length in meter.
    :param spectral_grid: numba dictionary containing the spectral grid (frequencies/directions)
    :param parameters: numba dictionary containing the parameters of the st4 algorithm. See the st4 class for details.
    :param wind_source: optional numpy array that contains the return values. This helps to avoid reallocating memory
        if the routine is called often in successive fashion (e.g. during iterative solutions).

    :return: numpy array of shape (nf,nd) containing for each frequency/direction pair the wind input in
        (m**2/Hz/Degree/s)
    """

    # Get the spectral size
    number_of_frequencies, number_of_directions = variance_density.shape

    vonkarman_constant = parameters["vonkarman_constant"]
    radian_frequency = spectral_grid["radian_frequency"]

    wind_forcing, wind_direction_degrees, wind_forcing_type = wind

    cosine_mutual_angle_wind_waves = cos(
        (spectral_grid["radian_direction"] - wind_direction_degrees * pi / 180 + pi)
        % (2 * pi)
        - pi
    )

    # Allocate the output variable
    if wind_source is None:
        wind_source = empty((number_of_frequencies, number_of_directions))

    # Preprocess winds to the correct conventions (U10->friction velocity if need be, wind direction degrees->radians)
    if wind_forcing_type == "u10":
        friction_velocity = (
            wind_forcing
            * vonkarman_constant
            / log(parameters["elevation"] / roughness_length)
        )
    elif wind_forcing_type in ["ustar", "friction_velocity"]:
        friction_velocity = wind_forcing

    else:
        raise ValueError("Unknown wind input type")

    # precalculate the constant multiplication.
    constant_factor = (
        parameters["growth_parameter_betamax"]
        / vonkarman_constant**2
        * parameters["air_density"]
        / parameters["water_density"]
    )

    if depth == inf:
        wavenumber = radian_frequency**2 / parameters["gravitational_acceleration"]
    else:
        wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)

    # wave age tuning times the directional cosine.
    cosine_wave_age_tuning = (
        parameters["wave_age_tuning_parameter"] * cosine_mutual_angle_wind_waves
    )

    # Loop over all frequencies/directions
    for frequency_index in range(number_of_frequencies):

        # Since wavenumber and wavespeed only depend on frequency we calculate those outside of the direction loop.
        relative_speed = (
            wavenumber[frequency_index]
            * friction_velocity
            / radian_frequency[frequency_index]
        )

        for direction_index in range(number_of_directions):
            # If the wave direction has an along wind component we continue.
            if cosine_mutual_angle_wind_waves[direction_index] > 0:

                # Calculate the directional factor in the wind input formulation
                W = relative_speed * cosine_mutual_angle_wind_waves[direction_index]

                # The logarithm of the approximate expression for the dimensionless critical height
                # (Ardhuin 2010, eq 20).
                log_dimensionless_critical_height = log(
                    wavenumber[frequency_index] * roughness_length
                ) + vonkarman_constant / (W + cosine_wave_age_tuning[direction_index])

                # if the dimensionless critical height > 1 we assume there is no more energy transfer (Janssen 1991),
                # his equation 18.
                if log_dimensionless_critical_height > 0:
                    log_dimensionless_critical_height = 0

                growth_rate = (
                    constant_factor
                    * exp(log_dimensionless_critical_height)
                    * log_dimensionless_critical_height**4
                    * W**2
                    * radian_frequency[frequency_index]
                )

                wind_source[frequency_index, direction_index] = (
                    growth_rate * variance_density[frequency_index, direction_index]
                )

            else:
                wind_source[frequency_index, direction_index] = 0.0

    return wind_source



from typing import Tuple

from numba import prange, jit
from numba_progress import ProgressBar
from numba.typed import List as NumbaList
from numpy import inf, nan, empty, isnan, zeros
from numpy.typing import NDArray

from roguewave.wavephysics.balance.dissipation import _bulk_dissipation_direction_point
from roguewave.wavephysics.balance._numba_settings import (
    numba_nocache_parallel,
    numba_nocache,
    numba_default,
)
from roguewave.wavephysics.balance.stress import (
    _roughness_estimate_point,
    _total_stress_point,
)
from roguewave.wavephysics.balance.solvers import numba_newton_raphson
from roguewave.wavespectra.operations import numba_integrate_spectral_data
from roguewave.wavephysics.balance.balance import SourceTermBalance
from xarray import DataArray, Dataset
from roguewave import FrequencyDirectionSpectrum


def windspeed_and_direction_from_spectra(
    balance: SourceTermBalance,
    guess_u10: DataArray,
    spectrum: FrequencyDirectionSpectrum,
    jacobian=False,
    jacobian_parameters=None,
    time_derivative_spectrum: FrequencyDirectionSpectrum = None,
    direction_iteration=False,
) -> Dataset:
    """

    :param bulk_rate:
    :param guess_u10:
    :param guess_direction:
    :param spectrum:
    :return:
    """
    disable = spectrum.number_of_spectra < 100
    if time_derivative_spectrum is None:
        time_derivative_spectrum = zeros(spectrum.shape())
    else:
        time_derivative_spectrum = time_derivative_spectrum.variance_density.values

    with ProgressBar(
        total=spectrum.number_of_spectra,
        disable=disable,
        desc=f"Estimating U10 from {balance.generation.name} and {balance.dissipation.name} "
        f"wind and dissipation source terms",
    ) as progress_bar:
        if not jacobian:
            speed, direction = _u10_from_spectra(
                variance_density=spectrum.variance_density.values,
                guess_u10=guess_u10.values,
                depth=spectrum.depth.values,
                wind_source_term_function=balance.generation._wind_source_term_function,
                tail_stress_parametrization_function=balance.generation._tail_stress_parametrization_function,
                dissipation_source_term_function=balance.dissipation._dissipation_function,
                parameters_generation=balance.generation.parameters,
                parameters_dissipation=balance.dissipation.parameters,
                spectral_grid=balance.generation.spectral_grid(spectrum),
                progress_bar=progress_bar,
                time_derivative_spectrum=time_derivative_spectrum,
                direction_iteration=direction_iteration,
            )
        else:
            if jacobian_parameters is None:
                raise ValueError(
                    "If gradients are requested a parameter list is required"
                )

            speed, direction, grad = _u10_from_spectra_gradient(
                variance_density=spectrum.variance_density.values,
                guess_u10=guess_u10.values,
                depth=spectrum.depth.values,
                wind_source_term_function=balance.generation._wind_source_term_function,
                tail_stress_parametrization_function=balance.generation._tail_stress_parametrization_function,
                dissipation_source_term_function=balance.dissipation._dissipation_function,
                grad_parameters=NumbaList(jacobian_parameters),
                parameters_generation=balance.generation.parameters,
                parameters_dissipation=balance.dissipation.parameters,
                spectral_grid=balance.generation.spectral_grid(spectrum),
                time_derivative_spectrum=time_derivative_spectrum,
                progress_bar=progress_bar,
                direction_iteration=direction_iteration,
            )
            grad = DataArray(data=grad)

    u10 = DataArray(
        data=speed,
        dims=spectrum.dims_space_time,
        coords=spectrum.coords_space_time,
    )

    direction = DataArray(
        data=direction,
        dims=spectrum.dims_space_time,
        coords=spectrum.coords_space_time,
    )

    if jacobian:
        return Dataset(data_vars={"u10": u10, "direction": direction, "jacobian": grad})
    else:
        return Dataset(data_vars={"u10": u10, "direction": direction})


@jit(**numba_nocache)
def _u10_from_bulk_rate_point(
    bulk_rate,
    variance_density,
    guess_u10,
    guess_direction,
    depth,
    spectral_grid,
    parameters,
    wind_source_term_function,
    tail_stress_parametrization_function,
    time_derivative_spectrum,
    direction_iteration,
):
    """

    :param bulk_rate:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param guess_u10:
    :param guess_direction:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :return:
    """

    direction = guess_direction
    u10 = guess_u10

    if bulk_rate == 0.0:
        return 0.0, guess_direction

    # We are tryomg to solve for U10; ~ accuracy of 0.01m/s is sufficient. We do not really care about relative accuracy
    # - for low winds (<1m/s) answers are really inaccurate anyway
    atol = 1.0e-2
    rtol = 1.0
    numerical_stepsize = 1e-3

    if direction_iteration:
        niter = 20
    else:
        niter = 1

    for ii in range(0, niter):
        wind_guess = (u10, direction, "u10")
        roughness_memory = [-1.0]
        args = (
            roughness_memory,
            variance_density,
            wind_guess,
            depth,
            wind_source_term_function,
            tail_stress_parametrization_function,
            spectral_grid,
            parameters,
            bulk_rate,
            time_derivative_spectrum,
        )

        try:
            u10 = numba_newton_raphson(
                _u10_iteration_function,
                u10,
                args,
                (0, inf),
                atol=atol,
                rtol=rtol,
                numerical_stepsize=numerical_stepsize,
            )
        except:
            u10 = nan
            break

        if direction_iteration:
            wind_guess = (u10, direction, "u10")
            _, new_direction = _total_stress_point(
                roughness_memory[0],
                variance_density,
                wind_guess,
                depth,
                wind_source_term_function,
                tail_stress_parametrization_function,
                spectral_grid,
                parameters,
            )

            direction_delta = (new_direction - direction + 180) % 360 - 180

            if abs(direction_delta) < 1.0:
                direction = new_direction
                break
            elif abs(direction_delta) < 10.0:
                direction = new_direction
            else:
                # Some under-relaxation
                direction = (direction + 0.5 * direction_delta) % 360

    return u10, direction


@jit(**numba_nocache)
def _u10_iteration_function(
    u10,
    memory_list,
    variance_density,
    wind_guess,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    spectral_grid,
    parameters,
    bulk_rate,
    time_derivative_spectrum,
):
    """
    To find the 10 meter wind that generates a certain bulk input we need to solve the inverse function for the
    wind input term. We define a function

            F(U10) = wind_bulk_rate(u10) - specified_bulk_rate

    and use a zero finder to solve for

            F(U10) = 0.

    This function defines the function F. The zero finder is responsible for calling it with the correct trailing
    arguments.

    :param wind_forcing: Wind input U10 we want to solve for.
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param wind_guess:
    :param depth:
    :param spectral_grid:
    :param parameters:
    :param bulk_rate:
    :return:
    """
    if u10 == 0.0:
        return -bulk_rate

    wind = (u10, wind_guess[1], wind_guess[2])

    # Estimate the rougness length
    roughness_length = _roughness_estimate_point(
        memory_list[0],
        variance_density,
        wind,
        depth,
        wind_source_term_function,
        tail_stress_parametrization_function,
        spectral_grid,
        parameters,
    )
    memory_list[0] = roughness_length

    # Calculate the wind input source term values
    generation = wind_source_term_function(
        variance_density,
        wind,
        depth,
        roughness_length,
        spectral_grid,
        parameters,
    )

    # Calculate the contribution of dEdt integrated over the spectrum. We only include spectral values in region where
    # there is a positive wind input
    bulk_time_derivative_spectrum = spectral_time_derivative_in_active_region(
        time_derivative_spectrum,
        generation,
        spectral_grid,
    )

    # Integrate the input and return the difference of the current guess with the desired bulk rate
    return (
        numba_integrate_spectral_data(generation, spectral_grid)
        - bulk_rate
        - bulk_time_derivative_spectrum
    )


@jit(**numba_default)
def spectral_time_derivative_in_active_region(
    time_derivative_spectrum: NDArray,
    generation: NDArray,
    spectral_grid,
):
    frequency_step = spectral_grid["frequency_step"]
    direction_step = spectral_grid["direction_step"]
    number_of_directions = time_derivative_spectrum.shape[1]
    number_of_frequencies = time_derivative_spectrum.shape[0]

    bulk_time_derivative_spectrum = 0.0

    for frequency_index in range(number_of_frequencies):
        for direction_index in range(number_of_directions):
            if generation[frequency_index, direction_index] > 0.0:
                bulk_time_derivative_spectrum += (
                    time_derivative_spectrum[frequency_index, direction_index]
                    * direction_step[direction_index]
                    * frequency_step[frequency_index]
                )
    return bulk_time_derivative_spectrum


@jit(**numba_nocache)
def _u10_from_spectra_point(
    variance_density,
    guess_u10,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    dissipation_source_term_function,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    time_derivative_spectrum,
    direction_iteration,
) -> Tuple[float, float]:
    """

    :param bulk_rate:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param guess_u10:
    :param depth:
    :param parameters:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """

    direction, bulk_rate = _bulk_dissipation_direction_point(
        variance_density,
        depth,
        dissipation_source_term_function,
        spectral_grid,
        parameters_dissipation,
    )

    # Note dissipation is negatve- but our target bulk wind generation is positive
    u10, direction = _u10_from_bulk_rate_point(
        -bulk_rate,
        variance_density,
        guess_u10,
        direction,
        depth,
        spectral_grid,
        parameters_generation,
        wind_source_term_function,
        tail_stress_parametrization_function,
        time_derivative_spectrum,
        direction_iteration,
    )
    return u10, direction


# ----------------------------------------------------------------------------------------------------------------------
# Functions for training (gradients with regard to parameter)
# ----------------------------------------------------------------------------------------------------------------------


@jit(**numba_nocache)
def _u10_parameter_gradient(
    variance_density,
    guess_u10,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    dissipation_source_term_function,
    grad_parameters,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    time_derivative_spectrum,
    direction_iteration,
) -> (float, float, NDArray):
    """
    Function to numerically calculate gradients for the requested coeficients
    :param bulk_rate:
    :param variance_density: The wave spectrum in m**2/Hz/deg
    :param guess_u10:
    :param depth:
    :param parameters:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """

    # Calculate the zero point
    direction, bulk_rate = _bulk_dissipation_direction_point(
        variance_density,
        depth,
        dissipation_source_term_function,
        spectral_grid,
        parameters_dissipation,
    )

    # Note dissipation is negatve- but our target bulk wind generation is positive
    u10, direction = _u10_from_bulk_rate_point(
        -bulk_rate,
        variance_density,
        guess_u10,
        direction,
        depth,
        spectral_grid,
        parameters_generation,
        wind_source_term_function,
        tail_stress_parametrization_function,
        time_derivative_spectrum,
        direction_iteration,
    )
    grad = empty(len(grad_parameters))
    if isnan(u10):
        grad[:] = 0
        return u10, direction, grad

    for index, param in enumerate(grad_parameters):
        perturbed_parameters_dissipation = parameters_dissipation.copy()
        perturbed_parameters_generation = parameters_generation.copy()

        if param in parameters_dissipation:
            step = 0.05 * abs(perturbed_parameters_dissipation[param])
            perturbed_parameters_dissipation[param] += step

            dissipation = dissipation_source_term_function(
                variance_density=variance_density,
                depth=depth,
                spectral_grid=spectral_grid,
                parameters=perturbed_parameters_dissipation,
            )

            new_bulk_rate = numba_integrate_spectral_data(dissipation, spectral_grid)

        else:
            step = 0.05 * abs(perturbed_parameters_generation[param])
            perturbed_parameters_generation[param] += step
            new_bulk_rate = bulk_rate

        new_u10, direction = _u10_from_bulk_rate_point(
            -new_bulk_rate,
            variance_density,
            u10,
            direction,
            depth,
            spectral_grid,
            perturbed_parameters_generation,
            wind_source_term_function,
            tail_stress_parametrization_function,
            time_derivative_spectrum,
            direction_iteration,
        )
        if isnan(new_u10):
            grad[index] = 0
        else:
            grad[index] = (new_u10 - u10) / step

    return u10, direction, grad


# ----------------------------------------------------------------------------------------------------------------------
# Apply to all spatial points
# ----------------------------------------------------------------------------------------------------------------------


@jit(**numba_nocache_parallel)
def _u10_from_spectra_gradient(
    variance_density,
    guess_u10,
    depth,
    wind_source_term_function,
    tail_stress_parametrization_function,
    dissipation_source_term_function,
    grad_parameters,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    time_derivative_spectrum,
    progress_bar: ProgressBar = None,
    direction_iteration=False,
) -> Tuple[NDArray, NDArray, NDArray]:
    """

    :param variance_density:
    :param guess_u10:
    :param depth:
    :param wind_source_term_function:
    :param dissipation_source_term_function:
    :param parameters_generation:
    :param parameters_dissipation:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """
    number_of_points = variance_density.shape[0]
    u10 = empty((number_of_points))
    direction = empty((number_of_points))
    grad = empty((number_of_points, len(grad_parameters)))
    for point_index in prange(number_of_points):
        if progress_bar is not None:
            progress_bar.update(1)

        (
            u10[point_index],
            direction[point_index],
            grad[point_index, :],
        ) = _u10_parameter_gradient(
            variance_density[point_index, :, :],
            guess_u10[point_index],
            depth[point_index],
            wind_source_term_function,
            tail_stress_parametrization_function,
            dissipation_source_term_function,
            grad_parameters,
            parameters_generation,
            parameters_dissipation,
            spectral_grid,
            time_derivative_spectrum[point_index, :, :],
            direction_iteration,
        )
    return u10, direction, grad


@jit(**numba_nocache_parallel)
def _u10_from_spectra(
    variance_density: NDArray,
    guess_u10: NDArray,
    depth: NDArray,
    wind_source_term_function,
    tail_stress_parametrization_function,
    dissipation_source_term_function,
    parameters_generation,
    parameters_dissipation,
    spectral_grid,
    progress_bar: ProgressBar = None,
    time_derivative_spectrum=None,
    direction_iteration=False,
) -> Tuple[NDArray, NDArray]:
    """

    :param variance_density:
    :param guess_u10:
    :param depth:
    :param wind_source_term_function:
    :param dissipation_source_term_function:
    :param parameters_generation:
    :param parameters_dissipation:
    :param spectral_grid:
    :param progress_bar:
    :return:
    """
    number_of_points = variance_density.shape[0]
    u10 = empty((number_of_points))
    direction = empty((number_of_points))

    for point_index in prange(number_of_points):
        if progress_bar is not None:
            progress_bar.update(1)

        u10[point_index], direction[point_index] = _u10_from_spectra_point(
            variance_density[point_index, :, :],
            guess_u10[point_index],
            depth[point_index],
            wind_source_term_function,
            tail_stress_parametrization_function,
            dissipation_source_term_function,
            parameters_generation,
            parameters_dissipation,
            spectral_grid,
            time_derivative_spectrum[point_index, :, :],
            direction_iteration,
        )
    return u10, direction



"""
Contents: Routines to calculate (inverse) linear dispersion relation and some related quantities such as phase and
group velocity. NOTE: the effect of surface currents is currently not included in these calculations.

The implementation uses numba to speed up calculations. Consequently, all functions are compiled to machine code, but
the first call to a function will be slow. Subsequent calls will be much faster.

Copyright (C) 2023
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Functions:
- `intrinsic_dispersion_relation`, calculate angular frequency for a given wavenumber and depth
- `inverse_intrinsic_dispersion_relation`, calculate wavenumber for a given angular frequency and depth
- `intrinsic_group_velocity`, calculate the group velocity given wave number and depth
- `phase_velocity`, calculate the phase velocity given wave number and depth
- `ratio_of_group_to_phase_velocity`, calculate the ratio of group to phase velocity given wave number and depth
- `jacobian_wavenumber_to_radial_frequency`, calculate the Jacobian of the wavenumber to radial frequency transformation
- `jacobian_radial_frequency_to_wavenumber`, calculate the Jacobian of the radial frequency to wavenumber transformation
"""

import numpy as np
from numba import jit
from ._tools import atleast_1d
from .constants import GRAV
from ._numba_settings import numba_default
from numbers import Real
from typing import Union


@jit(**numba_default)
def inverse_intrinsic_dispersion_relation(
        angular_frequency: Union[Real,np.ndarray],
        dep: Union[Real,np.ndarray],
        grav:Real=GRAV,
        maximum_number_of_iterations:int=10,
        tolerance:Real=1e-3,
) -> np.ndarray:
    """
    Find wavenumber k for a given radial frequency w using Newton Iteration.
    Exit when either maximum number of iterations is reached, or tolerance
    is achieved. Typically only 1 to 2 iterations are needed.

    :param w: radial frequency
    :param dep: depth in meters
    :param grav:  gravitational acceleration
    :param maximum_number_of_iterations: maximum number of iterations
    :param tolerance: relative accuracy
    :return: The wavenumber as a numpy array.
    """

    # Numba does not recognize "atleast_1d" for scalars
    w = atleast_1d(angular_frequency)

    k_deep_water_estimate = w ** 2 / grav
    k_shallow_water_estimate = w / np.sqrt(grav * dep)

    # == FIRST GUESS==
    # Use the intersection between shallow and deep water estimates to guestimate
    # which relation to use
    wavenumber_estimate = np.where(
        w > np.sqrt(grav / dep), k_deep_water_estimate, k_shallow_water_estimate
    )

    # == Newton Iteration ==
    error = intrinsic_dispersion_relation(wavenumber_estimate, dep, grav) - w
    for ii in range(0, maximum_number_of_iterations):
        # Calculate the derivative of the error function with respect to wavenumber. To note: the derivative is
        # merely the group velocity.
        kd = wavenumber_estimate * dep
        error_derivative_to_wavenumber = np.where(
            kd > 5,
            0.5 * w / wavenumber_estimate,
            (1 / 2 + kd / np.sinh(2 * kd)) * w / wavenumber_estimate
        )
        # Newton Iteration
        wavenumber_estimate = wavenumber_estimate - error / error_derivative_to_wavenumber

        # Update error
        error = intrinsic_dispersion_relation(wavenumber_estimate, dep, grav) - w

        # Check for convergence
        relative_absolute_error = np.abs(error) / w
        if np.all(relative_absolute_error < tolerance):
            break
    else:
        print('inverse_intrinsic_dispersion_relation:: No convergence in solving for wavenumber')

    return wavenumber_estimate


@jit(**numba_default)
def intrinsic_dispersion_relation(k, dep, grav=GRAV) -> np.ndarray:
    """
    The intrinsic dispersion relation for linear waves in water of constant depth that relates the specific angular
    frequency to a given wavenumber and depth in a reference frame following mean ambient flow.

    Wavenumber may be a scalar or a numpy array. The function always returns a numpy array. If depth is specified as a
    numpy array it must have the same shape as the wavenumber array.

    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return: Intrinsic angular frequency (rad/s)
    """
    k = atleast_1d(k)
    return np.sqrt(grav * k * np.tanh(k * dep))


@jit(**numba_default)
def phase_velocity(k, depth, grav=GRAV) -> np.ndarray:
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return intrinsic_dispersion_relation(k, depth, grav=grav) / k


@jit(**numba_default)
def ratio_group_velocity_to_phase_velocity(k, depth, grav) -> np.ndarray:
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    kd = k * depth
    return np.where(kd > 5, 0.5, 0.5 + kd / np.sinh(2 * kd))


@jit(**numba_default)
def intrinsic_group_velocity(k, depth, grav=GRAV) -> np.ndarray:
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return ratio_group_velocity_to_phase_velocity(k, depth, grav=grav) * phase_velocity(
        k, depth, grav
    )


@jit(**numba_default)
def jacobian_wavenumber_to_radial_frequency(k, depth, grav=GRAV) -> np.ndarray:
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return 1 / intrinsic_group_velocity(k, depth, grav)


@jit(**numba_default)
def jacobian_radial_frequency_to_wavenumber(k, depth, grav=GRAV) -> np.ndarray:
    """
    :param k: Wavenumber (rad/m)
    :param depth: Depth (m)
    :param grav: Gravitational acceleration (m/s^2)
    :return:
    """
    return intrinsic_group_velocity(k, depth, grav)


# Aliasses based on common notation in linear wave theory
c = phase_velocity
cg = intrinsic_group_velocity
k = inverse_intrinsic_dispersion_relation
w = intrinsic_dispersion_relation
n = ratio_group_velocity_to_phase_velocity



from pandas import DataFrame
from roguewave.interpolate.geometry import Geometry
from .timebase import TimeSlice
from roguewave.modeldata.open_remote import open_remote_dataset
from typing import Dict,List, Union
from roguewave.interpolate.dataset import interpolate_dataset

def extract_from_remote_dataset(
        geometry:Geometry,
        variable:Union[List[str],str],
        time_slice:TimeSlice,
        model_name:str,
        cache_name=None
) -> Dict[str, DataFrame]:
    """
    Extract timeseries at specified locations from the dataset identified by
    the model name and the timeslice.

    :param geometry: space(time) geometry we want to download from (e.g. a set
        of points, or a spotter track). For most use cases can be specified as

            geometry  =   ( [latitudes ] , [longitudes] )

        more advanced slicing will require the use of interpolation geometries
        defined in roguewave.interpolation.geometry.

    :param variable: name of the variable of interest. Can be a list in which
            case all variables in the list are retrieved.
    :param time_slice: time slice of interest.
    :param model_name: model name
    :param cache_name: name of local cache. If None, default cache setup will
        be used.

    :return: Dictionary where each geometry element is a key, and the value is
        a pandas dataframe with as index the time, and as columns the
        requested variables along the given geometry.
    """

    dataset = open_remote_dataset( variable, time_slice, model_name,
                                       cache_name=cache_name)
    return interpolate_dataset(dataset,geometry)





from numba import jit
from numpy import (
    log,
    inf,
    array,
    exp,
    empty,
    pi,
    cos,
    tanh,
    sqrt,
    linspace,
    sin,
    zeros,
    concatenate,
    min,
    trapz,
)

from roguewave.wavephysics.balance.solvers import numba_newton_raphson
from ._numba_settings import numba_default


@jit(**numba_default)
def tail_stress_parametrization_jb23(
    variance_density,
    wind,
    depth,
    roughness_length,
    spectral_grid,
    parameters,
):
    vonkarman_constant = parameters["vonkarman_constant"]
    radian_direction = spectral_grid["radian_direction"]
    elevation = parameters["elevation"]

    number_of_frequencies, number_of_directions = variance_density.shape
    direction_step = spectral_grid["direction_step"]

    wind_forcing, wind_direction_degrees, wind_forcing_type = wind
    wind_direction_radian = wind_direction_degrees * pi / 180
    cosine_mutual_angle = cos(radian_direction - wind_direction_radian)
    cosine = cos(radian_direction)
    sine = sin(radian_direction)

    if wind_forcing_type == "u10":
        friction_velocity = (
            wind_forcing * vonkarman_constant / log(elevation / roughness_length)
        )

    elif wind_forcing_type in ["ustar", "friction_velocity"]:
        friction_velocity = wind_forcing

    else:
        raise ValueError("Unknown wind input type")

    directional_integral = 0.0
    directional_integral_last_bin = 0.0
    directional_integral_last_bin_east = 0.0
    directional_integral_last_bin_north = 0.0
    for direction_index in range(0, number_of_directions):
        if cosine_mutual_angle[direction_index] <= 0.0:
            continue
        directional_integral += (
            variance_density[number_of_frequencies - 1, direction_index]
            * direction_step[direction_index]
        )

        directional_integral_last_bin += (
            cosine_mutual_angle[direction_index] ** 2
            * variance_density[number_of_frequencies - 1, direction_index]
            * direction_step[direction_index]
        )

        directional_integral_last_bin_east += (
            cosine_mutual_angle[direction_index] ** 2
            * cosine[direction_index]
            * variance_density[number_of_frequencies - 1, direction_index]
            * direction_step[direction_index]
        )

        directional_integral_last_bin_north += (
            cosine_mutual_angle[direction_index] ** 2
            * sine[direction_index]
            * variance_density[number_of_frequencies - 1, direction_index]
            * direction_step[direction_index]
        )

    if directional_integral_last_bin > 0.0:
        stress_east_fac = (
            directional_integral_last_bin_east / directional_integral_last_bin
        )
        stress_north_fac = (
            directional_integral_last_bin_north / directional_integral_last_bin
        )
    else:
        stress_east_fac = 1.0
        stress_north_fac = 0.0

    last_resolved_wavenumber = (
        spectral_grid["radian_frequency"][number_of_frequencies - 1] ** 2
        / parameters["gravitational_acceleration"]
    )

    jacobian_to_wavenumber_density = (
        spectral_grid["radian_frequency"][number_of_frequencies - 1]
        / last_resolved_wavenumber
        / 4
        / pi
    )

    starting_energy_wavenumber_density = (
        directional_integral_last_bin * jacobian_to_wavenumber_density
    )
    wavenumbers = wavenumber_grid(
        last_resolved_wavenumber, roughness_length, friction_velocity, parameters
    )

    if wavenumbers[0] < 0:
        return 0.0, 0.0

    saturation_spectrum = saturation_spectrum_parametrization(
        wavenumbers,
        starting_energy_wavenumber_density,
        last_resolved_wavenumber,
        friction_velocity,
        parameters,
    )

    background_stress = (
        parameters["charnock_constant"] ** 2
        * friction_velocity**6
        / parameters["gravitational_acceleration"] ** 2
        / roughness_length**2
    )

    tail_spectrum = saturation_spectrum * wavenumbers ** (-3)
    stress = wind_stress_tail(
        wavenumbers, roughness_length, friction_velocity, tail_spectrum, parameters
    )
    integral = (
        trapz(stress, wavenumbers) + background_stress * parameters["air_density"]
    )

    eastward_stress = integral * stress_east_fac

    northward_stress = integral * stress_north_fac

    return eastward_stress, northward_stress


# --------------------------------
# Janssen and Bidlot 2022
# --------------------------------
@jit(**numba_default)
def wind_stress_tail(
    wavenumbers,
    roughness_length,
    friction_velocity,
    tail_spectrum,
    parameters,
):

    windinput = wind_input_tail(
        wavenumbers, roughness_length, friction_velocity, tail_spectrum, parameters
    )

    angular_frequency = dispersion(
        wavenumbers,
        parameters["gravitational_acceleration"],
        parameters["surface_tension"],
    )

    stress = empty(wavenumbers.shape[0])
    for wavenumber_index in range(0, wavenumbers.shape[0]):
        stress[wavenumber_index] = (
            angular_frequency[wavenumber_index] * windinput[wavenumber_index]
        )

    return stress * parameters["water_density"]


# ----
# Helper functions Integration domains
# ----
@jit(**numba_default)
def log_bounds_wavenumber(roughness_length, friction_velocity, parameters):
    """
    Find the lower bound of the integration domain for JB2022.

    :param friction_velocity:
    :param effective_charnock:
    :param vonkarman_constant:
    :param wave_age_tuning_parameter:
    :param gravitational_acceleration:
    :return:
    """

    args = (roughness_length, friction_velocity, parameters)

    if friction_velocity <= 0.0:
        return array((-inf, -inf))

    # Wavenumber where miles_mu has a minimum value.
    miles_max_val_wavenumber = log(
        parameters["vonkarman_constant"] ** 2
        * parameters["gravitational_acceleration"]
        / 4
        / friction_velocity**2
    )

    if (
        miles_mu_cutoff(
            miles_max_val_wavenumber, roughness_length, friction_velocity, parameters
        )
        > 0.0
    ):
        # Not solvable. Essentially zero stress interval.
        return array((-inf, -inf))

    # find the right root
    log_upper_bound = numba_newton_raphson(
        miles_mu_cutoff,
        log(1.1 / roughness_length),
        args,
        (miles_max_val_wavenumber, log(1 / roughness_length)),
        verbose=False,
        name="log bound wavenumber 1",
    )

    guess = miles_max_val_wavenumber - 1

    # find the left root
    while True:
        log_lower_bound = numba_newton_raphson(
            miles_mu_cutoff,
            guess,
            args,
            (-inf, miles_max_val_wavenumber),
            verbose=False,
            name="log bound wavenumber 3",
        )
        if log_lower_bound < log_upper_bound * 0.98:
            break
        else:
            guess = guess - 1

    return array([log_lower_bound, log_upper_bound])


# ----
# Helper functions growth parameter
# ----
@jit(**numba_default)
def miles_mu(log_wavenumber, roughness_length, friction_velocity, parameters):
    vonkarman_constant = parameters["vonkarman_constant"]
    wave_age_tuning_parameter = parameters["wave_age_tuning_parameter"]
    gravitational_acceleration = parameters["gravitational_acceleration"]
    surface_tension = parameters["surface_tension"]

    wavenumber = exp(log_wavenumber)
    wavespeed = celerity(wavenumber, gravitational_acceleration, surface_tension)

    return (
        log_wavenumber
        + log(roughness_length)
        + vonkarman_constant
        / (friction_velocity / wavespeed + wave_age_tuning_parameter)
    )


@jit(**numba_default)
def miles_mu_cutoff(log_wavenumber, roughness_length, friction_velocity, parameters):
    return miles_mu(
        log_wavenumber, roughness_length, friction_velocity, parameters
    ) - log(1 + 0.25 * tanh(4 * friction_velocity**4))


@jit(**numba_default)
def wind_input_tail(
    wavenumbers, roughness_length, friction_velocity, tail_spectrum, parameters
):
    vonkarman_constant = parameters["vonkarman_constant"]
    growth_parameter_betamax = parameters["growth_parameter_betamax"]

    non_linear_effect_strength = parameters["non_linear_effect_strength"]

    number_of_wavenumbers = wavenumbers.shape[0]
    windinput = empty(number_of_wavenumbers)

    mu = miles_mu(log(wavenumbers), roughness_length, friction_velocity, parameters)

    epsilon = parameters["air_density"] / parameters["water_density"]

    wave_speed = celerity(
        wavenumbers,
        parameters["gravitational_acceleration"],
        parameters["surface_tension"],
    )  # parameters["surface_tension"])
    angular_frequency = dispersion(
        wavenumbers,
        parameters["gravitational_acceleration"],
        parameters["surface_tension"],
    )

    miles_cutoff = log(1 + 0.25 * tanh(4 * friction_velocity**4))
    for wavenumber_index in range(0, number_of_wavenumbers):
        if mu[wavenumber_index] > miles_cutoff:
            windinput[wavenumber_index] = 0.0
            continue

        linear_growth_parameter = (
            angular_frequency[wavenumber_index]
            * mu[wavenumber_index] ** 4
            * exp(mu[wavenumber_index])
            * growth_parameter_betamax
            / vonkarman_constant**2
            * epsilon
            * friction_velocity**2
            / wave_speed[wavenumber_index] ** 2
        )

        N2 = (
            non_linear_effect_strength
            * tail_spectrum[wavenumber_index]
            * linear_growth_parameter
            * wavenumbers[wavenumber_index] ** 2
            / vonkarman_constant
            / epsilon
            / friction_velocity
        )
        N1 = N2 / 6.0
        nonlinear_correction = (1.0 + N1) / (1.0 + N2)
        growth_parameter = linear_growth_parameter * nonlinear_correction

        windinput[wavenumber_index] = growth_parameter * tail_spectrum[wavenumber_index]

    return windinput


# ----
# Helper functions spectral parametrization tail
# ----


@jit(**numba_default)
def wavenumber_grid(
    starting_wavenumber, roughness_length, friction_velocity, parameters
):
    log_starting_wavenumber = log(starting_wavenumber)
    log_bounds = log_bounds_wavenumber(roughness_length, friction_velocity, parameters)

    if log_bounds[0] < log_starting_wavenumber:
        log_bounds[0] = log_starting_wavenumber

    if log_bounds[1] < log_starting_wavenumber:
        # Numba does not handle multiple returns well. Here I return a len 1 array with
        # a negative wavenumber to signal that there is no valid wavenumber grid
        return array([-1.0])

    # After Lenain and Melville, 2017
    log_upper_bound_eq_range = log(
        upper_limit_wavenumber_equilibrium_range(friction_velocity, parameters)
    )

    # Starting wave number where we switch on 3-wave interactions.
    log_three_wave_start = log(
        three_wave_starting_wavenumber(friction_velocity, parameters)
    )

    has_eq_range = log_bounds[0] < log_upper_bound_eq_range
    has_constant_range = (
        log_bounds[0] < log_three_wave_start
        and log_bounds[1] > log_upper_bound_eq_range
    )
    has_cap_range = log_bounds[1] > log_three_wave_start

    if has_eq_range:
        high = min(array((log_upper_bound_eq_range, log_bounds[1])))
        low = log_bounds[0]
        wavenumber = linspace(low, high, 50)
    else:
        wavenumber = zeros((1,))
        wavenumber[0] = log_bounds[0]

    if has_constant_range:
        high = min(array((log_three_wave_start, log_bounds[1])))
        low = wavenumber[-1]
        wavenumber = concatenate((wavenumber[:-1], linspace(low, high, 50)))

    if has_cap_range:
        high = log_bounds[1]
        low = wavenumber[-1]
        wavenumber = concatenate((wavenumber[:-1], linspace(low, high, 50)))

    return exp(wavenumber)


@jit(**numba_default)
def saturation_spectrum_parametrization(
    wavenumbers,
    energy_at_starting_wavenumber,
    starting_wavenumber,
    friction_velocity,
    parameters,
):
    """
    Saturation spectrum accordin to the VIERS model (adapted from JB2023)

    :param wavenumbers: set of wavenumbers
    :param energy_at_starting_wavenumber: variance density as a function of wavenumber,
        scaled such that int(e(k) dk = variance. This varies from Peter's work who uses an energy E such that e = E*k
        with k the wavenumber which originates from a transfer to polar coordinates of the 2d wavenumber spectrum.

    :param gravitational_acceleration: gravitational
    :param surface_tension:
    :param friction_velocity:
    :return:
    """

    gravitational_acceleration = parameters["gravitational_acceleration"]
    surface_tension = parameters["surface_tension"]

    # After Lenain and Melville, 2017
    upper_bound_eq_range = 0.01 * gravitational_acceleration / friction_velocity**2

    # Starting wave number where we switch on 3-wave interactions.
    three_wave_start = three_wave_starting_wavenumber(friction_velocity, parameters)

    number_of_wavenumbers = wavenumbers.shape[0]
    saturation_spectrum = empty(number_of_wavenumbers)

    # Saturation in the "saturation range", we assume a k**-3 spectrum here (f**-5)
    saturation_at_start_of_eq_range = (
        energy_at_starting_wavenumber * starting_wavenumber**3
    )

    #
    if starting_wavenumber < upper_bound_eq_range:
        saturation_at_end_of_eq_range = saturation_at_start_of_eq_range * sqrt(
            upper_bound_eq_range / starting_wavenumber
        )

    else:
        saturation_at_end_of_eq_range = saturation_at_start_of_eq_range

    # Strength of the 3-wave interactin parameter. This is directly taken from the VIERS work - as it was not
    # specified what was used in JB23.
    strength_three_wave_interactions = (
        3 * pi / 16 * (tanh(2 * (sqrt(wavenumbers / three_wave_start) - 1)) + 1)
    )

    # Strength at the point where we turn on the interactions
    strength_three_wave_interactions_start = 3 * pi / 16

    energy_flux_at_boundary = (
        strength_three_wave_interactions
        * saturation_at_end_of_eq_range**2
        * celerity(three_wave_start, gravitational_acceleration, surface_tension) ** 4
        / group_velocity(three_wave_start, gravitational_acceleration, surface_tension)
    )

    if surface_tension > 0.0:
        k0 = sqrt(gravitational_acceleration / surface_tension)
    else:
        k0 = inf

    c0 = (gravitational_acceleration * surface_tension) ** (1 / 4)
    for wavenumber_index in range(number_of_wavenumbers):
        if wavenumbers[wavenumber_index] < upper_bound_eq_range:
            # In the eq. range the saturation spectrum goes as sqrt(k)
            saturation_spectrum[
                wavenumber_index
            ] = saturation_at_start_of_eq_range * sqrt(
                wavenumbers[wavenumber_index] / starting_wavenumber
            )

        elif (
            wavenumbers[wavenumber_index] >= upper_bound_eq_range
            and wavenumbers[wavenumber_index] < three_wave_start
        ):
            # Constant saturation region
            saturation_spectrum[wavenumber_index] = saturation_at_end_of_eq_range

        elif wavenumbers[wavenumber_index] >= three_wave_start:
            # Region where three-wave interactions play a role
            scaling_constant = sqrt(
                energy_flux_at_boundary[wavenumber_index]
                / 2
                / strength_three_wave_interactions_start  # strength_three_wave_interactions[wavenumber_index]
            ) * c0 ** (-3 / 2)

            y = wavenumbers[wavenumber_index] / k0
            saturation_spectrum[wavenumber_index] = (
                scaling_constant
                * y
                * sqrt(1 + 3 * y**2)
                / ((1 + y**2) * (y + y**3) ** (1 / 4))
            )

    return saturation_spectrum


@jit(**numba_default)
def upper_limit_wavenumber_equilibrium_range(friction_velocity, parameters):
    """
    Upper limit eq. range
    :param gravitational_acceleration:
    :param surface_tension:
    :param friction_velocity:
    :return:
    """

    return 0.01 * parameters["gravitational_acceleration"] / friction_velocity**2


@jit(**numba_default)
def three_wave_starting_wavenumber(friction_velocity, parameters):
    """
    Starting wavenumber for the capilary-gravity part. See JB2023, eq 41 and 42.
    :param gravitational_acceleration:
    :param surface_tension:
    :param friction_velocity:
    :return:
    """
    if parameters["surface_tension"] == 0.0:
        return inf
    return (
        sqrt(parameters["gravitational_acceleration"] / parameters["surface_tension"])
        * 1
        / (1.48 + 2.05 * friction_velocity)
    )


# ----
# Helper functions grav. cap. waves
# ----
@jit(**numba_default)
def dispersion(wavenumber, gravitational_acceleration, surface_tension):
    return sqrt(
        gravitational_acceleration * wavenumber + surface_tension * wavenumber**3
    )


@jit(**numba_default)
def celerity(wavenumber, gravitational_acceleration, surface_tension):
    return sqrt(gravitational_acceleration / wavenumber + surface_tension * wavenumber)


@jit(**numba_default)
def group_velocity(wavenumber, gravitational_acceleration, surface_tension):
    return (
        1.0
        / 2.0
        * (gravitational_acceleration + 3 * surface_tension * wavenumber**2)
        / sqrt(
            gravitational_acceleration * wavenumber + surface_tension * wavenumber**3
        )
    )



_nopython = True
_nogil = False
_cache = False
_forceobj = False
_parallel = False
_error_model = "python"
_fastmath = False
_boundscheck = False

numba_default = {
    "nopython": _nopython,
    "nogil": _nogil,
    "cache": _cache,
    "forceobj": _forceobj,
    "parallel": _parallel,
    "error_model": _error_model,
    "fastmath": _fastmath,
    "boundscheck": _boundscheck,
}

numba_parallel = numba_default.copy()
numba_parallel["parallel"] = True

numba_nocache = numba_default.copy()
numba_nocache["cache"] = False

numba_nocache_parallel = numba_default.copy()
numba_nocache_parallel["cache"] = False
numba_nocache_parallel["parallel"] = True



from roguewave.wavephysics.balance.jb23_tail_stress import (
    tail_stress_parametrization_jb23,
)
from roguewave.wavephysics.balance.st4_wind_input import (
    ST4WindInput,
)
from roguewave.wavephysics.fluidproperties import GRAVITATIONAL_ACCELERATION, AIR, WATER

from numpy import cos, pi, log, exp, empty, inf, sin
from numba import njit
from roguewave.wavetheory import inverse_intrinsic_dispersion_relation
from typing import TypedDict


class JB23WaveGenerationParameters(TypedDict):
    gravitational_acceleration: float
    charnock_maximum_roughness: float
    charnock_constant: float
    air_density: float
    water_density: float
    vonkarman_constant: float
    wave_age_tuning_parameter: float
    growth_parameter_betamax: float
    elevation: float
    air_viscosity: float
    viscous_stress_parameter: float
    surface_tension: float
    width_factor: float
    non_linear_effect_strength: float


class JB23WindInput(ST4WindInput):
    name = "JB23 generation"

    def __init__(self, parameters: JB23WaveGenerationParameters = None):
        super(JB23WindInput, self).__init__(parameters)
        # The obly difference with ST4 as implemented is the calculation of the tail contributions.
        self._wind_source_term_function = _jb23_wind_generation_point
        self._tail_stress_parametrization_function = tail_stress_parametrization_jb23

    @staticmethod
    def default_parameters() -> JB23WaveGenerationParameters:
        return JB23WaveGenerationParameters(
            wave_age_tuning_parameter=0.008,
            growth_parameter_betamax=2,
            gravitational_acceleration=GRAVITATIONAL_ACCELERATION,
            charnock_maximum_roughness=inf,
            charnock_constant=0.005,
            air_density=AIR.density,
            water_density=WATER.density,
            vonkarman_constant=AIR.vonkarman_constant,
            elevation=10,
            air_viscosity=AIR.kinematic_viscosity,
            viscous_stress_parameter=1.0 / 25.0,
            surface_tension=WATER.kinematic_surface_tension,
            width_factor=0.6,
            non_linear_effect_strength=1.0,
        )


# ----------------------------------------------------------------------------------------------------------------------
# ST4 Point wise implementation functions (apply to a single spatial point)
# ----------------------------------------------------------------------------------------------------------------------
@njit(cache=True, fastmath=True)
def _jb23_wind_generation_point(
    variance_density,
    wind,
    depth,
    roughness_length,
    spectral_grid,
    parameters,
    wind_source=None,
):
    """
    Implememtation of the st4 wind growth rate term

    :param variance_density:
    :param wind:
    :param depth:
    :param roughness_length:
    :param spectral_grid:
    :param parameters:
    :return:
    """

    # Get the spectral size
    number_of_frequencies, number_of_directions = variance_density.shape

    # Allocate the output variable
    if wind_source is None:
        wind_source = empty((number_of_frequencies, number_of_directions))

    # Unpack variables passed in dictionaries/tuples for ease of reference and some slight speed benifits
    # (avoid dictionary lookup in loops).
    vonkarman_constant = parameters["vonkarman_constant"]
    growth_parameter_betamax = parameters["growth_parameter_betamax"]
    air_density = parameters["air_density"]
    water_density = parameters["water_density"]
    gravitational_acceleration = parameters["gravitational_acceleration"]
    radian_frequency = spectral_grid["radian_frequency"]
    radian_direction = spectral_grid["radian_direction"]
    direction_step = spectral_grid["direction_step"]
    elevation = parameters["elevation"]
    non_linear_effect_strength = parameters["non_linear_effect_strength"]

    wind_forcing, wind_direction_degrees, wind_forcing_type = wind

    # Preprocess winds to the correct conventions (U10->friction velocity if need be, wind direction degrees->radians)
    if wind_forcing_type == "u10":
        friction_velocity = (
            wind_forcing * vonkarman_constant / log(elevation / roughness_length)
        )

    elif wind_forcing_type in ["ustar", "friction_velocity"]:
        friction_velocity = wind_forcing

    else:
        raise ValueError("Unknown wind input type")

    wind_direction_radian = wind_direction_degrees * pi / 180

    # precalculate the constant multiplication.
    constant_factor = (
        growth_parameter_betamax / vonkarman_constant**2 * air_density / water_density
    )

    if depth == inf:
        wavenumber = radian_frequency**2 / gravitational_acceleration
    else:
        wavenumber = inverse_intrinsic_dispersion_relation(radian_frequency, depth)

    mutual_angle = (radian_direction - wind_direction_radian + pi) % (2 * pi) - pi
    cosine = cos(mutual_angle)
    sine2 = sin(mutual_angle) ** 2

    # Loop over all frequencies/directions
    epsilon = air_density / water_density

    cosine_wave_age_tuning = parameters["wave_age_tuning_parameter"] * cosine
    for frequency_index in range(number_of_frequencies):

        # Since wavenumber and wavespeed only depend on frequency we calculate those outside of the direction loop.
        wavespeed = radian_frequency[frequency_index] / wavenumber[frequency_index]
        group_speed = wavespeed / 2

        relative_speed = friction_velocity / wavespeed

        N1 = 0.0
        N2 = 0.0

        for direction_index in range(number_of_directions):
            # If the wave direction has an along wind component we continue.
            if cosine[direction_index] > 0:

                # Calculate the directional factor in the wind input formulation
                W = relative_speed * cosine[direction_index]

                effective_wave_age = log(
                    wavenumber[frequency_index] * roughness_length
                ) + vonkarman_constant / (W + cosine_wave_age_tuning[direction_index])

                if effective_wave_age > 0:
                    effective_wave_age = 0

                growth_rate = (
                    constant_factor
                    * exp(effective_wave_age)
                    * effective_wave_age**4
                    * W**2
                    * radian_frequency[frequency_index]
                )

                # Note that Peter does not include the k factor into the defenition of his F(k,\theta). So we need
                # an additional division by k (in addition to the jacobian cg) to make this work-
                # hence the wavenumber **2 !!
                common_factor = (
                    non_linear_effect_strength
                    * variance_density[frequency_index, direction_index]
                    * group_speed
                    * direction_step[direction_index]
                    * wavenumber[frequency_index] ** 2
                    * growth_rate
                    / vonkarman_constant
                    / friction_velocity
                    / epsilon
                    / 2
                    / pi
                )

                N1 += common_factor * sine2[direction_index]
                N2 += common_factor

                wind_source[frequency_index, direction_index] = (
                    growth_rate * variance_density[frequency_index, direction_index]
                )
            else:
                wind_source[frequency_index, direction_index] = 0.0

        #
        for direction_index in range(number_of_directions):
            wind_source[frequency_index, direction_index] = (
                wind_source[frequency_index, direction_index] * (1 + N1) / (1 + N2)
            )

    return wind_source



import numpy
from numba import njit
from roguewave.interpolate.spline import cubic_spline

from roguewave.tools.grid import midpoint_rule_step
from roguewave.tools.time_integration import integrated_response_factor_spectral_tail
from scipy.interpolate import CubicSpline


@njit(cache=True)
def _fit(x, y, power):
    X = x**power
    coef = numpy.sum(X * y) / numpy.sum(X * X)
    goodness_of_fit = numpy.sum((coef * X - y) ** 2)
    return coef, goodness_of_fit


@njit(cache=True)
def tail_fit(x, y, power):
    if power is None:
        coef_4, goodness_of_fit_4 = _fit(x, y, -4)
        coef_5, goodness_of_fit_5 = _fit(x, y, -5)
        if goodness_of_fit_5 < goodness_of_fit_4:
            return coef_5, -5
        else:
            return coef_4, -4

    else:
        coef, goodness_of_fit = _fit(x, y, power)
        return coef, power


@njit(cache=True)
def transition_frequency(
    energy_last_resoved_frequency, last_resolved_frequency, integrated_tail
):
    """
    Estimate the transition frequency from a -4 to -5 tail given the last resolved frequency and the integrated energy
    in the tail.

    :param energy_last_resoved_frequency:
    :param last_resolved_frequency:
    :param integrated_tail:
    :return:
    """
    return (
        (energy_last_resoved_frequency * last_resolved_frequency**4)
        / (
            4 * energy_last_resoved_frequency * energy_last_resoved_frequency
            - 12 * integrated_tail
        )
    ) ** (1 / 3)


@njit(cache=True)
def numba_fill_zeros_or_nan_in_tail(
    variance_density: numpy.ndarray,
    frequencies: numpy.ndarray,
    power=None,
    zero=0.0,
    points_in_fit=10,
    tail_information=None,
):
    input_shape = variance_density.shape
    number_of_elements = 1
    for value in input_shape[:-1]:
        number_of_elements *= value

    number_of_frequencies = input_shape[-1]
    variance_density = variance_density.reshape((number_of_elements, input_shape[-1]))

    if tail_information is None:
        tail_energy = numpy.zeros(number_of_elements)
        tail_bounds = (frequencies[-1] + 1, frequencies[-1] + 2)
    else:
        tail_energy = tail_information[1]
        tail_bounds = tail_information[0]

    for ii in range(0, number_of_elements):

        for jj in range(number_of_frequencies - 1, -1, -1):
            if variance_density[ii, jj] > zero:
                # Note, for nan values this is also always false. No need to seperately check for that.
                index = jj
                break
        else:
            # no valid value found, we cannot extrapolate. Technically this is not needed as we catch this below as
            # well (since index=number_of_frequencies by default). But to make more explicit that we catch for this
            # scenario I leave it in.
            continue

        if index == number_of_frequencies - 1:
            continue
        elif index < points_in_fit:
            continue

        last_resolved_frequency = frequencies[index]
        last_resolved_energy = variance_density[ii, index]

        if tail_energy[ii] > 0.0:
            trans_freq, starting_energy = _compound_tail(
                last_resolved_frequency,
                last_resolved_energy,
                tail_energy[ii],
                tail_bounds,
            )

            for jj in range(index + 1, number_of_frequencies):
                if frequencies[jj] >= trans_freq:
                    variance_density[ii, jj] = (
                        starting_energy * trans_freq * frequencies[jj] ** -5
                    )
                else:
                    variance_density[ii, jj] = starting_energy * (frequencies[jj]) ** -4

        else:
            coef, fitted_power = tail_fit(
                x=frequencies[index - points_in_fit + 1 : index + 1],
                y=variance_density[ii, index - points_in_fit + 1 : index + 1],
                power=power,
            )

            for jj in range(index + 1, number_of_frequencies):
                variance_density[ii, jj] = coef * (frequencies[jj]) ** fitted_power

    return numpy.reshape(variance_density, input_shape)


@njit(cache=True)
def _compound_tail(
    last_resolved_frequency, last_resolved_energy, raw_tail_energy, tail_bounds
):
    transition_frequencies = numpy.linspace(tail_bounds[0], tail_bounds[1], 11)

    goodness_of_fit = 0.0
    transition_frequency = transition_frequencies[0]
    starting_energy = 0.0
    for index in range(transition_frequencies.shape[0]):
        tail_energy = raw_tail_energy * integrated_response_factor_spectral_tail(
            -4,
            tail_bounds[0],
            tail_bounds[1],
            2.5,
            transition_frequency=transition_frequencies[index],
        )
        freq_int = 1.0 / 3.0 * (
            tail_bounds[0] ** -3 - transition_frequencies[index] ** -3
        ) + 1.0 / 4.0 * transition_frequencies[index] * (
            transition_frequencies[index] ** -4 - tail_bounds[1] ** -4
        )
        current_starting_energy = tail_energy / freq_int
        current_fit = (
            current_starting_energy * last_resolved_frequency**-4
            - last_resolved_energy
        ) ** 2

        if index == 0:
            goodness_of_fit = (
                current_starting_energy * last_resolved_frequency**-5
                - last_resolved_energy
            ) ** 2
            starting_energy = current_starting_energy
            transition_frequency = transition_frequencies[index]

        else:
            if current_fit < goodness_of_fit:
                goodness_of_fit = current_fit
                starting_energy = current_starting_energy
                transition_frequency = transition_frequencies[index]

    return transition_frequency, starting_energy


def _cdf_interpolate_spline(
    frequency: numpy.ndarray,
    frequency_spectrum: numpy.ndarray,
    monotone_interpolation: bool = False,
    frequency_axis=-1,
    **kwargs
) -> CubicSpline:
    """
    Interpolate the spectrum using the cdf.

    :param interpolation_frequency: frequencies to estimate the spectrum at.
    :param frequency: Frequencies of the spectrum. Shape = ( nf, )
    :param frequency_spectrum: Frequency Variance density spectrum. Shape = ( np , nf )
    :param interpolating_spline_order: Order of the spline to use in the interpolation (max 5 supported by scipy)
    :param positive: Ensure the output is positive (e.g. for A1 or B1 densities output need not be strictly positive).
    :return:
    """

    # Calculate the binsize for each of the frequency bins. We assume that the given frequencies represent the center
    # of a bin, and that the bin width at frequency i is determined as the sum of half the up and downwind differences:
    #
    #  frequency_step[i]   =  (frequency_step[i] - frequency_step[i-1])/2 + (frequency_step[i+1] - frequency_step[i])/2
    #
    # At boundaries we simply take twice the up or downwind difference, e.g.:
    #
    # frequency_step[0] = (frequency_step[1] - frequency_step[0])
    #
    frequency_step = midpoint_rule_step(frequency)
    integration_frequencies = numpy.concatenate(([0], numpy.cumsum(frequency_step)))
    integration_frequencies = (
        integration_frequencies - frequency_step[0] / 2 + frequency[0]
    )

    cumsum = numpy.cumsum(frequency_spectrum * frequency_step, axis=frequency_axis)
    shape = list(cumsum.shape)
    shape[frequency_axis] = 1
    cumsum = numpy.concatenate((numpy.zeros(shape), cumsum), axis=-1)

    interpolator = cubic_spline(
        integration_frequencies, cumsum, monotone_interpolation=monotone_interpolation
    )

    return interpolator


def spline_peak_frequency(
    frequency: numpy.ndarray,
    frequency_spectrum: numpy.ndarray,
    frequency_axis=-1,
    monotone_interpolation=True,
) -> numpy.ndarray:
    """
    Estimate the peak frequency of the spectrum based on a cubic spline interpolation of the partially integrated
    variance.

    :param frequency: Frequencies of the spectrum. Shape = ( nf, )
    :param frequency_spectrum: Frequency Variance density spectrum. Shape = ( np , nf )
    :return: peak frequencies. Shape = ( np, )
    """
    #

    interpolator = _cdf_interpolate_spline(
        frequency, frequency_spectrum, monotone_interpolation, frequency_axis
    )
    interpolator = interpolator.derivative()
    # Find the maxima of the density function by setting dEdf = 0, and finding the roots of all the splines representing
    # the density function.
    list_of_roots_for_all_spectra = interpolator.derivative().roots()

    if frequency_spectrum.ndim == 1:
        list_of_roots_for_all_spectra = [list_of_roots_for_all_spectra]

    # initialize output memory
    peak_frequency = numpy.zeros(len(list_of_roots_for_all_spectra))

    # We now have a list in which each entry contains all the roots for that given spectrum. Loop over each spectrum
    # and..
    for index, root in enumerate(list_of_roots_for_all_spectra):
        # ..evaluate density spectrum at those roots.
        values_at_roots = interpolator(root)

        # Because the interpolator returns values at the current roots evaluated at _all_ spectra, we still have to
        # select the values at the spectrum of interest. This implementation is silly, adds computational costs, and can
        # probably be improved. It seems "fast enough" so that I'll punt that to another time.
        if len(list_of_roots_for_all_spectra) > 1:
            values_at_roots = values_at_roots[index, :]

        _root = root[numpy.isfinite(values_at_roots)]
        values_at_roots = values_at_roots[numpy.isfinite(values_at_roots)]

        # ... get the root that corresponds to the largest peak.
        index_peak = numpy.argmax(values_at_roots)
        peak_frequency[index] = _root[index_peak]

    return peak_frequency



_nopython = True
_nogil = True
_cache = True
_forceobj = False
_parallel = False
_error_model = "python"
_fastmath = False
_boundscheck = False

numba_default = {
    "nopython": _nopython,
    "nogil": _nogil,
    "cache": _cache,
    "forceobj": _forceobj,
    "parallel": _parallel,
    "error_model": _error_model,
    "fastmath": _fastmath,
    "boundscheck": _boundscheck,
}



from numba import njit
from numpy import log, exp, array, pi, cos, sin

from roguewave.wavephysics.balance.solvers import numba_newton_raphson

#
# ----
# Wam Cy47r1 Implementation
# ----
#
# IFS DOCUMENTATION – Cy47r1 Operational implementation 30 June 2020 - PART VII
#


@njit()
def log_dimensionless_critical_height(
    x, charnock_constant, vonkarman_constant, wave_age_tuning_parameter
):
    """

    Dimensionless Critical Height according to Janssen (see IFS Documentation).
    :param x:
    :param charnock_constant:
    :param vonkarman_constant:
    :param wave_age_tuning_parameter:
    :return:
    """
    return (
        log(charnock_constant)
        + 2 * x
        + vonkarman_constant / (exp(x) + wave_age_tuning_parameter)
    )


@njit()
def integrate_tail_frequency_distribution(
    lower_bound, effective_charnock, vonkarman_constant, wave_age_tuning_parameter
):
    """
    Integrate the tail of the distributions. We are integrating

    log(Z(Y))Z(Y)**4 / Y      for   Y0 <= Y <= 1

    where

    Y = u_* / wavespeed
    Z = charnock * Y**2 * exp( vonkarman_constant / ( Y + wave_age_tuning_parameter)

    The boundaries of the integral are defined as the point where the critical height is at the surface
    (Y=1) and the point where Z >= 1 ( Y = Y0).

    We follow section 5 in the WAM documentation (see below). And introduce x = log(Y)

    so that we integrate in effect over

    log(Z(x))Z(x)**4     x0 <= x <= 0

    We find x0 as the point where Z(x0) = 0.

    REFERENCE:

    IFS DOCUMENTATION – Cy47r1 Operational implementation 30 June 2020 - PART VII

    :param effective_charnock:
    :param vonkarman_constant:
    :param wave_age_tuning_parameter:
    :return:
    """

    args = (effective_charnock, vonkarman_constant, wave_age_tuning_parameter)

    # find the location of the lower boundary of the integration domain. THis is where
    # loglog_mu = 0
    x0 = numba_newton_raphson(
        log_dimensionless_critical_height, log(0.01), args, (-10, 0), verbose=False
    )

    log_lower_bound = log(lower_bound)

    if log_lower_bound > x0:
        x0 = log_lower_bound

    if x0 > 0.0:
        return 0.0

    # Define the stepsize of the integration. Since the upper boundary is 0, this is merely the start
    stepsize = -x0 / 4

    # We use Boole's rule for integration.
    evaluation_points = array([x0, 3 * x0 / 4, 2 * x0 / 4, x0 / 4, 0])
    log_waveage = log_dimensionless_critical_height(evaluation_points, *args)
    values = log_waveage**4 * exp(log_waveage)
    integrant = (
        2
        / 45
        * stepsize
        * (
            7 * values[0]
            + 32 * values[1]
            + 12 * values[2]
            + 32 * values[3]
            + 7 * values[4]
        )
    )
    return integrant


@njit()
def tail_stress_parametrization_wam(
    variance_density,
    wind,
    depth,
    roughness_length,
    spectral_grid,
    parameters,
):
    vonkarman_constant = parameters["vonkarman_constant"]
    growth_parameter_betamax = parameters["growth_parameter_betamax"]
    wave_age_tuning_parameter = parameters["wave_age_tuning_parameter"]
    gravitational_acceleration = parameters["gravitational_acceleration"]
    radian_frequency = spectral_grid["radian_frequency"]
    radian_direction = spectral_grid["radian_direction"]
    elevation = parameters["elevation"]
    air_density = parameters["air_density"]

    number_of_frequencies, number_of_directions = variance_density.shape
    direction_step = spectral_grid["direction_step"]

    wind_forcing, wind_direction_degrees, wind_forcing_type = wind
    wind_direction_radian = wind_direction_degrees * pi / 180
    cosine_mutual_angle = cos(radian_direction - wind_direction_radian)
    cosine = cos(radian_direction)
    sine = sin(radian_direction)

    if wind_forcing_type == "u10":
        friction_velocity = (
            wind_forcing * vonkarman_constant / log(elevation / roughness_length)
        )

    elif wind_forcing_type in ["ustar", "friction_velocity"]:
        friction_velocity = wind_forcing

    else:
        raise ValueError("Unknown wind input type")

    directional_integral_last_bin_east = 0.0
    directional_integral_last_bin_north = 0.0
    for direction_index in range(0, number_of_directions):
        if cosine_mutual_angle[direction_index] <= 0.0:
            continue

        directional_integral_last_bin_east += (
            cosine_mutual_angle[direction_index] ** 2
            * cosine[direction_index]
            * variance_density[number_of_frequencies - 1, direction_index]
            * direction_step[direction_index]
        )

        directional_integral_last_bin_north += (
            cosine_mutual_angle[direction_index] ** 2
            * sine[direction_index]
            * variance_density[number_of_frequencies - 1, direction_index]
            * direction_step[direction_index]
        )

    effective_charnock = (
        roughness_length * gravitational_acceleration / friction_velocity**2
    )

    lower_bound = (
        friction_velocity
        * radian_frequency[number_of_frequencies - 1]
        / gravitational_acceleration
    )
    frequency_integral = integrate_tail_frequency_distribution(
        lower_bound, effective_charnock, vonkarman_constant, wave_age_tuning_parameter
    )

    constant = (
        radian_frequency[number_of_frequencies - 1] ** 5
        / (2 * pi * gravitational_acceleration**2)
        * friction_velocity**2
        * growth_parameter_betamax
        / vonkarman_constant**2
    )

    # Add contribution of cappiliary waves
    total_stress = parameters["air_density"] * friction_velocity**2
    charnock_roughness = _charnock_relation_point(friction_velocity, parameters)
    background_stress = charnock_roughness**2 / roughness_length**2 * total_stress

    eastward_stress = (
        directional_integral_last_bin_east * frequency_integral * constant * air_density
        + cos(wind_direction_radian) * background_stress
    )

    northward_stress = (
        directional_integral_last_bin_north
        * frequency_integral
        * constant
        * air_density
        + sin(wind_direction_radian) * background_stress
    )

    return eastward_stress, northward_stress


@njit(cache=True)
def _charnock_relation_point(friction_velocity, parameters):
    """
    Charnock relation
    :param friction_velocity:
    :param parameters:
    :return:
    """
    roughness_length = (
        friction_velocity**2
        / parameters["gravitational_acceleration"]
        * parameters["charnock_constant"]
    )

    if roughness_length > parameters["charnock_maximum_roughness"]:
        roughness_length = parameters["charnock_maximum_roughness"]

    return roughness_length



from roguewave.wavephysics.balance import WindGeneration, Dissipation
from typing import Mapping, Dict
from roguewave import FrequencyDirectionSpectrum
from xarray import DataArray


class SourceTermBalance:
    def __init__(self, generation: WindGeneration, disspipation: Dissipation):
        self.generation = generation
        self.dissipation = disspipation

    def evaluate_imbalance(
        self,
        wind_speed: DataArray,
        wind_direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        time_derivative_spectrum: FrequencyDirectionSpectrum = None,
    ) -> DataArray:

        if time_derivative_spectrum is None:
            time_derivative_spectrum = 0.0
        else:
            time_derivative_spectrum = time_derivative_spectrum.variance_density

        return (
            self.generation.rate(
                spectrum,
                wind_speed,
                wind_direction,
            )
            + self.dissipation.rate(spectrum)
            - time_derivative_spectrum
        )

    def evaluate_bulk_imbalance(
        self,
        wind_speed: DataArray,
        wind_direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        time_derivative_spectrum: FrequencyDirectionSpectrum = None,
    ) -> DataArray:

        if time_derivative_spectrum is None:
            time_derivative_spectrum = 0.0
        else:
            time_derivative_spectrum = time_derivative_spectrum.m0()

        return (
            self.generation.bulk_rate(spectrum, wind_speed, wind_direction)
            + self.dissipation.bulk_rate(spectrum)
            - time_derivative_spectrum
        )

    def update_parameters(self, parameters: Mapping):
        self.generation.update_parameters(parameters)
        self.dissipation.update_parameters(parameters)

    @property
    def get_parameters(self) -> Dict:
        return self.generation._parameters | self.dissipation._parameters



from roguewave import (
    FrequencyDirectionSpectrum,
)
from roguewave.wavephysics.balance.stress import (
    _roughness_estimate,
    _wave_supported_stress,
    _tail_supported_stress
)
from roguewave.wavespectra.operations import numba_integrate_spectral_data
from roguewave.wavephysics.balance.source_term import SourceTerm
from xarray import DataArray, zeros_like, Dataset
from typing import Literal
from numpy import empty, sqrt, log

from roguewave.wavephysics.balance._numba_settings import (
    numba_default,
    numba_nocache_parallel,
)
from numba import jit, types, prange
from numba.typed import Dict as NumbaDict
from numpy.typing import NDArray
from typing import Mapping

TWindInputType = Literal["u10", "friction_velocity", "ustar"]


class WindGeneration(SourceTerm):
    name = "base"

    def __init__(self, parmaters):
        super(WindGeneration, self).__init__(parmaters)
        self._wind_source_term_function = None
        self._tail_stress_parametrization_function = _tail_stress_parametrization_none

    def update_parameters(self, parameters: Mapping):
        for key in parameters:
            if key in self._parameters:
                self._parameters[key] = parameters[key]

    def rate(
        self,
        spectrum: FrequencyDirectionSpectrum,
        speed: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> DataArray:

        if roughness_length is None:
            roughness_length = self.roughness(
                speed, direction, spectrum, wind_speed_input_type=wind_speed_input_type
            )

        wind = (speed.values, direction.values, wind_speed_input_type)
        wind_input = _wind_generation(
            spectrum.variance_density.values,
            wind=wind,
            depth=spectrum.depth.values,
            roughness_length=roughness_length.values,
            wind_source_term_function=self._wind_source_term_function,
            spectral_grid=self.spectral_grid(spectrum),
            parameters=self.parameters,
        )
        return DataArray(data=wind_input, dims=spectrum.dims, coords=spectrum.coords())

    def bulk_rate(
        self,
        spectrum: FrequencyDirectionSpectrum,
        speed: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> DataArray:
        wind = (speed.values, direction.values, wind_speed_input_type)

        if roughness_length is None:
            roughness_length = self.roughness(
                speed, direction, spectrum, wind_speed_input_type=wind_speed_input_type
            )

        wind_input = _bulk_wind_generation(
            spectrum.variance_density.values,
            wind,
            spectrum.depth.values,
            roughness_length.values,
            self._wind_source_term_function,
            self.spectral_grid(spectrum),
            self.parameters,
        )
        return DataArray(
            data=wind_input,
            dims=spectrum.dims_space_time,
            coords=spectrum.coords_space_time,
        )

    def roughness(
        self,
        speed: DataArray,
        direction: DataArray,
        spectrum: FrequencyDirectionSpectrum,
        roughness_length_guess: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> DataArray:
        if roughness_length_guess is None:
            roughness_length_guess = zeros_like(speed) - 1

        wind = (speed.values, direction.values, wind_speed_input_type)

        rougness = _roughness_estimate(
            guess=roughness_length_guess.values,
            variance_density=spectrum.variance_density.values,
            wind=wind,
            depth=spectrum.depth.values,
            wind_source_term_function=self._wind_source_term_function,
            tail_stress_parametrization_function=self._tail_stress_parametrization_function,
            spectral_grid=self.spectral_grid(spectrum),
            parameters=self.parameters,
        )
        return DataArray(
            data=rougness,
            dims=spectrum.dims_space_time,
            coords=spectrum.coords_space_time,
        )

    def stress(
        self,
        spectrum: FrequencyDirectionSpectrum,
        speed: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> Dataset:

        if roughness_length is None:
            roughness_length = self.roughness(
                speed, direction, spectrum, wind_speed_input_type=wind_speed_input_type
            )

        wind = (speed.values, direction.values, wind_speed_input_type)

        stress = _wave_supported_stress(
            variance_density=spectrum.variance_density.values,
            wind=wind,
            depth=spectrum.depth.values,
            roughness_length=roughness_length.values,
            wind_source_term_function=self._wind_source_term_function,
            tail_stress_parametrization_function=self._tail_stress_parametrization_function,
            spectral_grid=self.spectral_grid(spectrum),
            parameters=self.parameters,
        )

        return Dataset(
            data_vars={
                "stress": (spectrum.dims_space_time, stress[0]),
                "direction": (spectrum.dims_space_time, stress[1]),
            },
            coords=spectrum.coords_space_time,
        )

    def friction_velocity(
        self,
        spectrum: FrequencyDirectionSpectrum,
        u10: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
    ) -> Dataset:
        stress = self.stress(spectrum, u10, direction, roughness_length, "u10")
        return sqrt(stress["stress"] / self.parameters["air_density"])

    def drag(
        self,
        spectrum: FrequencyDirectionSpectrum,
        speed: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ):
        if wind_speed_input_type == "u10":
            u10 = speed
            u_star = self.friction_velocity(
                spectrum, speed, direction, roughness_length
            )
        else:
            u_star = speed
            roughness = self.roughness(
                speed, direction, spectrum, wind_speed_input_type=wind_speed_input_type
            )
            u10 = u_star * self.parameters["vonkarman_constant"] * log(10 / roughness)

        return DataArray(data=u_star.values**2 / u10.values**2)

    def tail_stress(
        self,
        spectrum: FrequencyDirectionSpectrum,
        speed: DataArray,
        direction: DataArray,
        roughness_length: DataArray = None,
        wind_speed_input_type: TWindInputType = "u10",
    ) -> Dataset:

        if roughness_length is None:
            roughness_length = self.roughness(
                speed, direction, spectrum, wind_speed_input_type=wind_speed_input_type
            )

        wind = (speed.values, direction.values, wind_speed_input_type)
        stress = _tail_supported_stress(
            variance_density=spectrum.variance_density.values,
            wind=wind,
            depth=spectrum.depth.values,
            roughness_length=roughness_length.values,
            tail_stress_parametrization_function=self._tail_stress_parametrization_function,
            spectral_grid=self.spectral_grid(spectrum),
            parameters=self.parameters,
        )

        return Dataset(
            data_vars={
                "stress": (spectrum.dims_space_time, stress[0]),
                "direction": (spectrum.dims_space_time, stress[1]),
            },
            coords=spectrum.coords_space_time,
        )


# ----------------------------------------------------------------------------------------------------------------------
# Apply to all spatial points
# ----------------------------------------------------------------------------------------------------------------------


@jit(**numba_nocache_parallel)
def _wind_generation(
    variance_density,
    wind,
    depth,
    roughness_length,
    wind_source_term_function,
    spectral_grid,
    parameters,
) -> NDArray:
    (
        number_of_points,
        number_of_frequencies,
        number_of_directions,
    ) = variance_density.shape

    generation = empty((number_of_points, number_of_frequencies, number_of_directions))
    for point_index in range(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])
        wind_generation = wind_source_term_function(
            variance_density[point_index, :, :],
            wind_at_point,
            depth[point_index],
            roughness_length[point_index],
            spectral_grid,
            parameters,
        )
        generation[point_index, :, :] = wind_generation
    return generation


@jit(**numba_nocache_parallel)
def _bulk_wind_generation(
    variance_density,
    wind,
    depth,
    roughness_length,
    wind_source_term_function,
    spectral_grid,
    parameters,
) -> NDArray:
    number_of_points = variance_density.shape[0]
    generation = empty((number_of_points))

    for point_index in prange(number_of_points):
        wind_at_point = (wind[0][point_index], wind[1][point_index], wind[2])
        wind_generation = wind_source_term_function(
            variance_density=variance_density[point_index, :, :],
            wind=wind_at_point,
            depth=depth[point_index],
            roughness_length=roughness_length[point_index],
            spectral_grid=spectral_grid,
            parameters=parameters,
        )
        generation[point_index] = numba_integrate_spectral_data(
            wind_generation, spectral_grid
        )
    return generation


# ----------------------------------------------------------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------------------------------------------------------


# ----------------------------------------------------------------------------------------------------------------------
# Helper functions
# ----------------------------------------------------------------------------------------------------------------------


@jit(**numba_default)
def _spectral_grid(radian_frequency, radian_direction, frequency_step, direction_step):
    return {
        "radian_frequency": radian_frequency,
        "radian_direction": radian_direction,
        "frequency_step": frequency_step,
        "direction_step": direction_step,
    }


def _numba_parameters(**kwargs):
    _dict = NumbaDict.empty(key_type=types.unicode_type, value_type=types.float64)
    for key in kwargs:
        _dict[key] = kwargs[key]

    return _dict


@jit(**numba_default)
def _tail_stress_parametrization_none(
    variance_density,
    wind,
    depth,
    roughness_length,
    spectral_grid,
    parameters,
):
    return 0.0



from typing import Mapping, Sequence, Union, Tuple
from pandas import DataFrame, concat
from xarray import DataArray
from datetime import timedelta

from roguewave.spotterapi.spotterapi import get_spotter_data
from roguewave.modeldata.extract import extract_from_remote_dataset
from roguewave.interpolate.dataframe import interpolate_dataframe_time
from roguewave.modeldata.timebase import TimeSlice


# =============================================================================
def colocate_model_spotter(variable: Union[Sequence[str], str], spotter_ids: Sequence[str], time_slice: TimeSlice,
                           model_name: str, cache_name: str = None, timebase: str = "model") -> Tuple[
    DataFrame, DataFrame
]:
    """
    Colocate spoter output and modek data
    :param variable: name of the variable of interest. Can be a list in which
            case all variables in the list are retrieved.
    :param spotter_ids: List of spotter ids of interest
    :param time_slice: time slice of interest.
    :param model_name: model name
    :param cache_name: name of local cache. If None, default cache setup will
        be used.
    :param timebase: Can be ['native','spotter','model'] if
        - native: output for spotters and model at spotters will be generated
                  at their own valid times
        - spotter: model data will be interpolated to spotter reporting time
        - model: Spotter data will be interpolated to model time
    :param return_as_dataset: Instead of a dictornary return everything as one
        xarray dataset ("Pandas on steroids"), where spotter_id is one of the
        dimensions of the datastructure (a "column" if you will)
    :return:
    """

    if timebase.lower() == "model":
        # If we have a model time base we interpolate spotter observations onto the model grid
        # To ensure we have data to use for interpolation at the end points we widen the interval there
        obs_delta = timedelta(hours=3)
    else:
        obs_delta = timedelta(hours=0)

    df = get_spotter_data(
        spotter_ids,
        "waves",
        time_slice.start_time - obs_delta,
        time_slice.end_time + obs_delta,
    )
    spotters = {
        key: value.drop(columns="spotter_id")
        for (key, value) in df.groupby("spotter_id")
    }

    for spotter_id in list(spotters.keys()):
        # If only one value - pop because we cannot interpolate on 1 value.
        if spotters[spotter_id].shape[0] <= 2:
            spotters.pop(spotter_id)

    # we want model data to cover the entire period. Spotter data will be returned up to the enddate. Hence we enforce
    # inclusion of the endpoint during colocation, this will ensure that whenever there is spotter data, there is
    # enclosing model data (which I would argue is the expected behaviour).
    if timebase.lower() == "spotter":
        time_slice.endpoint = True

    model = extract_from_remote_dataset(
        spotters,
        variable,
        time_slice,
        model_name,
        cache_name=cache_name,
    )

    for spotter_id in spotters:
        s = spotters[spotter_id]  # type: DataFrame
        m = model[spotter_id]  # type: DataFrame

        if timebase.lower() == "native":
            pass
        elif timebase.lower() == "observed" or timebase.lower() == "spotter":
            m = interpolate_dataframe_time(m, s["time"].values)
        elif timebase.lower() == "model":
            s = interpolate_dataframe_time(s, m["time"].values)
        else:
            raise ValueError(
                f"Unknown timebase {timebase}, must be one of: "
                f"native, observed, model, or spotter"
            )

        model[spotter_id] = m
        spotters[spotter_id] = s

    model = colocate_into_single_dataframe(model)
    spotters = colocate_into_single_dataframe(spotters)
    return model, spotters


def colocate_into_single_dataframe(data: dict[str, DataFrame]) -> DataFrame:
    _ids = list(data.keys())
    values = [data[_id] for _id in _ids]

    data = concat(values, keys=_ids, names=["spotter_id", "time index"])
    data.reset_index(inplace=True)
    data.drop(columns="time index", inplace=True)
    return data



import numpy

try:
    from qpsolvers import solve_ls
except ImportError:
    pass

from .utils import get_direction_increment, get_rhs, get_constraint_matrix


def log_likelyhood(
    directions_radians: numpy.ndarray,
    a1: numpy.ndarray,
    b1: numpy.ndarray,
    a2: numpy.ndarray,
    b2: numpy.ndarray,
    progress,
    **kwargs
) -> numpy.ndarray:
    number_of_frequencies = a1.shape[-1]
    number_of_points = a1.shape[0]

    directional_distribution = numpy.zeros(
        (number_of_points, number_of_frequencies, len(directions_radians))
    )

    for ipoint in range(0, number_of_points):
        progress.update(1)
        directional_distribution[ipoint, :, :] = _log_likelyhood(
            directions_radians,
            a1[
                ipoint,
                :,
            ],
            b1[ipoint, :],
            a2[ipoint, :],
            b2[ipoint, :],
        )
    return directional_distribution


def _log_likelyhood(directions_radians: numpy.ndarray, a1, b1, a2, b2) -> numpy.ndarray:
    """
    Return the directional distribution that minimizes the variance (D**2)
    constrained by given observed directional moments,

    :param directions_radians: 1d array of wave directions in radians,
    length[number_of_directions]

    :param a1: 1d array of cosine directional moment as function of frequency,
    length [number_of_frequencies]

    :param b1: 1d array of sine directional moment as function of frequency,
    length [number_of_frequencies]

    :param a2: 1d array of double angle cosine directional moment as function
    of frequency, length [number_of_frequencies]

    :param b2: 1d array of double angle sine directional moment as function of
    frequency, length [number_of_frequencies]

    :return: array with shape [number_of_frequencies,number_of_direction]
    representing the directional distribution of the waves at each frequency.

    Minimize the variance of the solution:

           integrate D**2 over directions

    such that the resulting distribution D reproduces the observed moments.

    Implementation notes:
    - we formulate the problem as a standard Quadratic Programming problem which
      can them be solved efficiently with the qpsolvers package.
    """

    a1 = numpy.atleast_1d(a1)
    b1 = numpy.atleast_1d(b1)
    a2 = numpy.atleast_1d(a2)
    b2 = numpy.atleast_1d(b2)

    number_of_frequencies = len(a1)
    directional_distribution = numpy.zeros(
        (number_of_frequencies, len(directions_radians))
    )

    constraint_matrix = get_constraint_matrix(directions_radians)
    rhs = get_rhs(a1, b1, a2, b2)
    identity_matrix = numpy.diag(numpy.ones_like(directions_radians), 0)

    zeros = numpy.zeros_like(directions_radians)
    direction_increment = get_direction_increment(directions_radians)
    upperbound = numpy.ones_like(directions_radians) / direction_increment.min()

    for ifreq in range(0, number_of_frequencies):
        res = solve_ls(
            R=identity_matrix,
            s=zeros,
            # minimizing |Rx-b|**2
            lb=zeros,
            ub=upperbound,
            # lb: non-negative; ub: binwidth * ub = 1
            A=constraint_matrix,
            b=rhs[ifreq, :],
            verbose=False
            # with hard constraint that Ax=b
        )

        if res is None:
            raise Exception("No solution")

        directional_distribution[ifreq, :] = res

    directional_distribution[directional_distribution < 0] = 0
    return numpy.squeeze(directional_distribution)
